# Copyright (c) 2013, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import os
import sys
import time
import stat

from workbench.tcp_utils import SocketClient
from workbench.os_utils import FileUtils, OSUtils
import subprocess

# The script is executed with elevated privileges which means it is 
# being executed on a clean environment on the admin user and so it is
# not possible to send back the command output
# 
# It uses a socket client to send the output to a command listener
# Which requires authentication (done with the handshake logic)
# and also a way to tell the listener that the command has completed

# Retrieves the parameters
params = ' '.join(sys.argv[1:])
port, handshake, done_key, command = params.split(' ', 3)

class CommandProcessor:
    def __init__(self, command, client):
        self._command, self._args = command.split(' ', 1)
        self._client = client
        self._result_code = 0
        self._result_message = ""
        
        self._commands = {}
        self._commands['LISTDIR'] = self._process_listdir
        self._commands['GETFILE'] = self._process_getfile
        self._commands['GETFILE_LINES'] = self._process_getfile_lines
        self._commands['GET_FREE_SPACE'] = self._process_get_free_space
        self._commands['CHECK_DIR_WRITABLE'] = self._process_check_dir_writable
        self._commands['CHECK_PATH_EXISTS'] = self._process_check_path_exists
        self._commands['CREATE_DIRECTORY'] = self._process_create_directory
        self._commands['CREATE_DIRECTORY_RECURSIVE'] = self._process_create_directory_recursive
        self._commands['REMOVE_DIRECTORY'] = self._process_remove_directory
        self._commands['REMOVE_DIRECTORY_RECURSIVE'] = self._process_remove_directory_recursive
        self._commands['DELETE_FILE'] = self._process_delete_file
        self._commands['COPY_FILE'] = self._process_copy_file
        self._commands['GET_FILE_OWNER'] = self._process_get_file_owner
        self._commands['EXEC'] = self._process_exec

        if not self._commands.has_key(self._command):
            raise RuntimeError("Command %s is not supported" % self._command)
    
    def execute(self):
        """
        Selects the proper method for the command execution,
        when an additional custom command (i.e. LISTDIR) is needed
        a method should be created on this class and called on this
        method based on the command name.

        When no specific method is defined for the given command it
        will be handled by the _execute_command method which assumes
        the command is valid for the operating system
        """
        self._commands[self._command]()
        
    def _process_listdir(self):
        """
        Lists the content of a directory and returns it to the command listener,
        either including or not the file sizes.
        Syntax:
                LISTDIR <size> <path>
                
                size:   Indicates if the returned list should containg the size
                        or not. 0 indicates no, 1 indicates yes.
                path:   The path to the file or folder that will be listed.
        """
        include_size, path = self._args.split(' ', 1)
        include_size = (include_size == '1')
        
        FileUtils.list_dir(path, include_size, self._send_to_listener)

    def _process_getfile(self):
        """
        Load data from a file and sends it back to the command listener in 64K chunks.
        Syntax:
                GETFILE <offset> <size> <path>
                
                offset: In bytes, indicates the position of the file where the
                        read operation will start. 0 indicates the beggining of
                        the file.
                size:   In bytes, indicates the amount of bytes to be read from
                        the file. 0 indicates the whole file should be read.
                path:   The path to the file that will be read.
        """
        offset, size, path = self._args.split(' ', 2)
        read_size = chunk_size = 64000
        
        offset = int(offset)
        size = int(size)
        remaining = size
        
        try:
            f = open(path, 'r')
        
            # If specifieds moves the read to the offset position
            if offset:
                f.seek(offset)
                
            # Starts the read process
            continue_reading = True
            while continue_reading:
                # Calculates the amount of data to be read, it will be the chunk size
                # most of the time, except when a specific amount of data is requested,
                # in such case the limit needs to be considered
                if size:
                    read_size = chunk_size if chunk_size < remaining else remaining
                
                data = f.read(read_size)
                
                # The read operation returns an empty string when no more data
                # is in the file
                if data:
                    self._client.send(data)
                    
                    # When a size is specified, the read will be done once that
                    # size is reached
                    if size:
                        remaining = remaining - read_size
                        
                        if remaining == 0:
                            continue_reading = False
                else:
                    continue_reading = False
            
            f.close()
                       
        except (IOError), e:
            self._result_code = 1
            self._result_message = repr(e)

    def _process_getfile_lines(self):
        #GETFILE_LINES <skip> <path>
        skip, path = self._args.split(' ', 1)
        skip = int(skip)
        FileUtils.get_file_lines(path, skip, self._send_to_listener)
        
    def _process_get_free_space(self):
        free_space = FileUtils.get_free_space(self._args)
        self._client.send(str(free_space))
            
    def _process_check_dir_writable(self):
        self._client.send(str(FileUtils.check_dir_writable(self._args)))
        
    def _process_check_path_exists(self):
        self._client.send(str(FileUtils.check_path_exists(self._args)))
        
    def _process_create_directory(self):
        FileUtils.create_directory(self._args)
        
    def _process_create_directory_recursive(self):
        FileUtils.create_directory_recursive(self._args)
        
    def _process_remove_directory(self):
        FileUtils.remove_directory(self._args)
        
    def _process_remove_directory_recursive(self):
        FileUtils.remove_directory_recursive(self._args)
        
    def _process_delete_file(self):
        FileUtils.delete_file(self._args)
        
    def _process_copy_file(self):
        files = self._args.split('>')
        
        if len(files) < 2:
            raise RuntimeError('Invalid call to the COPY_FILE command')
        
        src = files[0]
        tgt = files[1]
        tgt_backup = ""
        if len(files) > 2:
            tgt_backup = files[2]
            
        src.strip()
        tgt.strip()
        tgt_backup.strip()
        FileUtils.copy_file(src, tgt, tgt_backup)
        
    def _process_get_file_owner(self):
        self._client.send(str(FileUtils.get_file_owner(self._args)))
        
    # Process any valid OS command that has not been handled by the other functions
    def _send_to_listener(self, data):
        self._client.send(data)
        
    def _process_exec(self):
        """
        Executes any OS valid command and sends the output to the command listener
        Syntax:
                EXEC <command>
                
                command: A valid OS command.
        """
        OSUtils.exec_command(self._args, self._send_to_listener)


# Creates the client to use it to send the output to the command listener
client = SocketClient('127.0.0.1', int(port), handshake, done_key)
            
try:
    if client.start():

        processor = CommandProcessor(command, client)

        processor.execute()
        
        client.close(processor._result_code, processor._result_message)
    
except Exception, e:
    if client._connected:
        client.close(1, repr(e))
    else:
        print e

            


# Copyright (c) 2013, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

#import cairo
import cairo_utils

def draw_varrow(cr, tip_pos, ah = 4, aw = 4):
    cr.new_path()
    x, y = tip_pos
    cr.move_to(x, y)
    cr.line_to(x - aw/2, y + ah - 0.5)
    cr.line_to(x + aw/2, y + ah - 0.5)
    cr.close_path()


def draw_harrow(cr, tip_pos, ah = 4, aw = 4):
    cr.new_path()
    x, y = tip_pos
    cr.move_to(x, y)
    cr.line_to(x - ah - 0.5, y - aw/2)
    cr.line_to(x - ah - 0.5, y + aw/2)
    cr.close_path()

#

class Settings:
    default_font = "Helvetica"


settings = Settings()


class Canvas(object):
    def __init__(self, set_needs_repaint_cb):
        self._size = (0, 0)

        self.tooltip = None

        self.set_needs_repaint = set_needs_repaint_cb

        self._figures = []

        self._mouse_down_figures = [None] * 5
        self._hover_figure = None
        self._background_color = (1,1,1)


    def deactivate(self):
        if self.tooltip:
            self.tooltip.close()
            self.tooltip = None


    def activate(self):
        pass


    def set_background_color(self, r, g, b):
        self._background_color = r, g, b


    def add(self, obj):
        obj.set_canvas(self)
        self._figures.append(obj)
        obj.invalidate()


    def remove(self, obj):
        self.invalidate(*obj.bounds)
        self.set_needs_repaint()
        obj.set_canvas(None)
        self._figures.remove(obj)


    def repaint(self, cr, x, y, w, h):
        cr.set_source_rgb(*self._background_color)
        cr.rectangle(0, 0, w, h)
        cr.fill()
        cr.translate(x, y)
        for f in self._figures:
            f.relayout(cr)
            f.render(cr)


    def invalidate(self, x, y, w, h):
        self.set_needs_repaint(x, y, w, h)


    def figure_at(self, x, y):
        for f in self._figures:
            if f.contains_point(x, y):
                return f
        return None


    # event handling
    def mouse_down(self, button, x, y):
        fig = self.figure_at(x, y)
        if fig:
            self._mouse_down_figures[button] = fig
            fig.mouse_down(button, x, y)
            return True
        return False

    def mouse_up(self, button, x, y):
        if self._mouse_down_figures[button]:
            self._mouse_down_figures[button].mouse_up(button, x, y)
            self._mouse_down_figures[button] = None


    def mouse_leave(self):
        #if self.tooltip:
        #    self.tooltip.close()
        #    self.tooltip = None
        pass


    def mouse_move(self, x, y):
        dragged = False
        for b, f in enumerate(self._mouse_down_figures):
            if f:
                dragged = True
                f.mouse_drag(b, x, y)
                break
        if not dragged:
            fig = self.figure_at(x, y)
            if fig != self._hover_figure:
                if self._hover_figure:
                    self._hover_figure.hover_out(x, y)
                if fig:
                    fig.hover_in(x, y)
                self._hover_figure = fig
            else:
                if self._hover_figure:
                    self._hover_figure.hover(x, y)
            if self._hover_figure:
                self._hover_figure.mouse_move(x, y)

# Base Classes

class Element(object):
    def __init__(self):
        self.canvas = None
        self.parent = None
        self._layout_dirty = False
        self._color = (0, 0, 0, 1)
    
        self.on_hover_in = None
        self.on_hover_out = None

    def set_canvas(self, canvas):
        self.canvas = canvas

    def invalidate(self, repaint_only = False):
        if not repaint_only:
            self._layout_dirty = True
        if self.canvas:
            self.canvas.invalidate(*self.bounds)

    def relayout(self, cr):
        if self._layout_dirty:
            self.do_relayout(cr)
        self._layout_dirty = False

    def do_relayout(self, cr):
        pass

    def render(self, cr):
        pass

    def set_color(self, r, g, b, a = 1.0):
        self._color = (r, g, b, a)

    # event handling
    def mouse_down(self, button, x, y):
        pass

    def mouse_up(self, button, x, y):
        pass

    def mouse_drag(self, button, x, y):
        pass

    def mouse_move(self, x, y):
        pass
    
    def hover_in(self, x, y):
        if self.on_hover_in:
            self.on_hover_in(self, x, y)

    def hover_out(self, x, y):
        if self.on_hover_out:
            self.on_hover_out(self, x, y)

    def hover(self, x, y):
        pass



HFill = 1 << 0
VFill = 1 << 1


class Figure(Element):
    def __init__(self):
        Element.__init__(self)
        self._x = 0
        self._y = 0
        self._width = 0
        self._height = 0
        self._uwidth = None
        self._uheight = None
        self._fill_color = (1, 1, 1, 1)
        self._line_width = 1
        self._dash = (None, None)
        self._padding = (0, 0, 0, 0)
        self._layout_flags = HFill | VFill

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def root_x(self):
        return self.x + (self.parent.root_x if self.parent else 0)

    @property
    def root_y(self):
        return self.y + (self.parent.root_y if self.parent else 0)

    @property
    def width(self):
        return self._width if self._uwidth is None else self._uwidth

    @property
    def height(self):
        return self._height if self._uheight is None else self._uheight

    @property
    def pos(self):
        return (self._x, self._y)

    @property
    def size(self):
        return (self.width, self.height)

    @property
    def bounds(self):
        return 0, 0, self.width, self.height

    @property
    def frame(self):
        return self.x, self.y, self.width, self.height

    @property
    def padding_top(self):
        return self._padding[0]

    @property
    def padding_left(self):
        return self._padding[1]

    @property
    def padding_bottom(self):
        return self._padding[2]

    @property
    def padding_right(self):
        return self._padding[3]

    def move(self, x, y):
        if self.canvas:
            self.canvas.invalidate(*self.frame)
        self._x = x
        self._y = y
        if self.canvas:
            self.canvas.invalidate(*self.frame)

    def contains_point(self, x, y):
        pos = self.pos
        size = self.size
        return x >= pos[0] and x < pos[0] + size[0] and y >= pos[1] and y < pos[1] + size[1]
    
    def center(self):
        x, y = self.pos
        w, h = self.size
        return int(x + w/2), int(y + h/2)

    def top_vertex(self):
        x, y = self.pos
        w, h = self.size
        return (x, y), (x+w, y)

    def bottom_vertex(self):
        x, y = self.pos
        w, h = self.size
        return (x, y+h), (x+w, y+h)

    def right_vertex(self):
        x, y = self.pos
        w, h = self.size
        return (x+w, y), (x+w, y+h)

    def left_vertex(self):
        x, y = self.pos
        w, h = self.size
        return (x, y), (x, y+h)
    

    def set_padding(self, t, l, b, r):
        self._padding = t, l, b, r

    def set_layout_flags(self, flags):
        self._layout_flags = flags
    
    def set_line_dash(self, dash, offset):
        self._dash = (dash, offset)
        self.invalidate()

    def set_line_width(self, l):
        self._line_width = l
        self.invalidate()

    def set_fill_color(self, r, g, b, a = 1.0):
        self._fill_color = (r, g, b, a)
        self.invalidate()

    def apply_attributes(self, c):
        c.set_source_rgba(*self._color)
        if self._dash[0]:
            c.set_dash(self._dash[0], self._dash[1])
        c.set_line_width(self._line_width)

    def apply_fill_attributes(self, c):
        c.set_source_rgba(*self._fill_color)

    def set_usize(self, w, h):
        self._uwidth = w
        self._uheight = h


class Line(object):
    def __init__(self):
        Element.__init__(self)
        self._end1 = None
        self._end2 = None

    @property
    def bounds(self):
        return 0, 0, 0, 0

    def contains_point(self, x, y):
        return None


#
#

class Container(Figure):
    def __init__(self):
        Figure.__init__(self)
        self._items = []

    def add(self, item):
        item.parent = self
        self._items.append(item)


    def remove(self, item):
        item.parent = None
        self._items.remove(item)


    def render(self, cr):
        cr.save()
        cr.translate(self.x, self.y)
        for i in self._items:
            i.render(cr)
        cr.restore()


class VBoxFigure(Container):
    def __init__(self):
        Container.__init__(self)
        self._spacing = 0
    
    
    def set_spacing(self, sp):
        self._spacing = sp

    def do_relayout(self, cr):
        lp, tp, rp, bp = self._padding
        y = tp
        x = lp
        max_width = self._width
        for i in self._items:
            i.do_relayout(cr)
            i.move(x, y)
            max_width = max(max_width, i.width)
            y += int(i.height) + self._spacing
        if self._items:
            y -= self._spacing

        for i in self._items:
            if i._layout_flags & HFill:
                i.set_usize(max_width, i._uheight)
                i.do_relayout(cr)
            else:
                i.move((max_width - i.width) / 2, i.y)

        self._width = max_width + lp + rp
        self._height = y + bp

        self.invalidate()

#
#

class TextFigure(Figure):
    def __init__(self, text=""):
        Figure.__init__(self)
        self._text = text
        self._font = settings.default_font
        self._font_size = 12
        self._text_color = (0, 0, 0, 1)
        self._line_spacing = 2
        self._bold = False
        self._xalignment = 0.0
        self._yalignment = 0.0

        self._line_height = 14
        self._text_height = 0

    def set_text_color(self, r, g, b, a=1.0):
        self._text_color = r, g, b, a
        self.invalidate()
    
    def set_font_size(self, s):
        self._font_size = s

    def set_font_bold(self, s):
        self._bold = s

    def set_alignment(self, x, y):
        self._xalignment = x
        self._yalignment = y
        self.invalidate(False)


    def set_text(self, text):
        self._text = text
        self.invalidate(False)


    def do_relayout(self, ctx):
        ctx.save()
        ctx.set_font(self._font, False, self._bold)
        ctx.set_font_size(self._font_size)

        if "\n" in self._text:
            lines = self._text.split("\n")
            w, h = 0, 0
            lh = 0
            self._text_height = 0
            for line in lines:
                ext = ctx.text_extents(line)
                w = max(w, int(ext.x_bearing + ext.x_advance))
                lh = max(lh, int(ext.height + ext.height + ext.y_bearing))
                self._text_height += ext.height + self._line_spacing
            if lines:
                self._text_height -= self._line_spacing
            self._line_height = int(lh)
            t, l, b, r = self._padding
            
            self._width, self._height = (w + r+l, self._text_height + t+b)
        else:
            ext = ctx.text_extents(self._text)
            self._extents = ext
            t, l, b, r = self._padding
            self._line_height = int(ext.height) + int(ext.height + ext.y_bearing)
            self._width = int(ext.x_bearing + ext.x_advance) + r+l
            self._height = self._line_height + t+b
            self._text_height = self._line_height
        ctx.restore()


    def render(self, ctx):
        ctx.save()
        ctx.translate(self.x, self.y)
        ctx.set_source_rgba(*self._text_color)
        ctx.set_font(self._font, False, self._bold)
        ctx.set_font_size(self._font_size)
        t, l, b, r = self._padding
        w, h = self.size
        x = int(l) + 0.5
        y = int(t) + 0.5 + int((self.height - t - b - self._text_height) * self._yalignment)

        for line in self._text.split("\n"):
            extents = ctx.text_extents(line)
            ctx.move_to(x + int((self.width - l - r - extents.width) * self._xalignment),
                        y + int(extents.height - (extents.height + extents.y_bearing)))
            ctx.show_text(line)
            y += extents.height + self._line_spacing

        ctx.stroke()
        ctx.restore()


class ImageFigure(Figure):
    def __init__(self, file):
        Figure.__init__(self)
        
        self.set_line_width(1)

        # if this is a @2x image (for hidpi), we need to scale it by half when painting on screen
        if "@2x" in file:
            self._scale = 0.5
        else:
            self._scale = 1.0
        self._image = cairo_utils.ImageSurface.from_png(file)
        self._width = self._image.get_width() * self._scale
        self._height = self._image.get_height() * self._scale


    def set_image_file(self, path):
        self._image = cairo_utils.ImageSurface.from_png(path)
        self._width = self._image.get_width() * self._scale
        self._height = self._image.get_height() * self._scale


    def render(self, ctx):
        ctx.save()
        ctx.translate(self.x, self.y)
        if self._scale != 1.0:
            ctx.scale(self._scale, self._scale)

        ctx.set_source_surface(self._image, 0, 0)
        ctx.paint()
        
        ctx.restore()



class ShapeFigure(TextFigure):
    def __init__(self, caption):
        TextFigure.__init__(self, caption)
        
        self.set_alignment(0.5, 0.5)
        self.set_line_width(1)
    
    
    def render(self, ctx):
        ctx.save()
        ctx.translate(self.x, self.y)
        
        self.make_path(ctx)
        
        self.apply_fill_attributes(ctx)
        ctx.fill_preserve()
        self.apply_attributes(ctx)
        ctx.stroke()
        ctx.restore()

        TextFigure.render(self, ctx)


    def make_path(self, ctx):
        ctx.rectangle(*self.bounds)


class DiamondShapeFigure(ShapeFigure):
    def make_path(self, ctx):
        ctx.move_to(0, self.height/2)
        ctx.line_to(self.width/2, 0)
        ctx.line_to(self.width, self.height/2)
        ctx.line_to(self.width/2, self.height)
        ctx.close_path()


class RectangleShapeFigure(ShapeFigure):
    _corner_radius = 0
    
    def make_path(self, ctx):
        if self._corner_radius == 0:
            ctx.rectangle(0.5, 0.5, self.width, self.height)
        else:
            ctx.rounded_rect(0.5, 0.5, self.width, self.height, self._corner_radius)
    
    def set_corner_radius(self, r):
        self._corner_radius = r


#
#

class HoverTip(Figure):
    def __init__(self, text):
        Figure.__init__(self)
        self.text = text
        self.font_size = 12
        self.bold = False
        self.line_spacing = 4
        self.line_height = 10
        self.padding = (4, 4, 4, 4)


    def calc(self, ctx):
        ctx.save()
        ctx.set_font_size(self.font_size)
        if self.bold:
            ctx.set_font("Helvetica", False, self.bold)
        if "\n" in self.text:
            lines = self.text.split("\n")
            w, h = 0, 0
            lh = 0
            for line in lines:
                ext = ctx.text_extents(line)
                w = max(w, int(ext.x_bearing + ext.x_advance))
                lh = max(lh, int(ext.height + (ext.height + ext.y_advance + ext.y_bearing)))
            self.line_height = lh
            h = lh * len(lines) + self.line_spacing * (len(lines)-1)
            t, l, b, r = self.padding
            self._width, self._height = (w + r+l, h + t+b)
        else:
            ext = ctx.text_extents(self.text)
            self._extents = ext
            t, l, b, r = self.padding
            self.line_height = int(ext.height + (ext.height + ext.y_advance + ext.y_bearing))
            self._width = int(ext.x_bearing + ext.x_advance) + r+l
            self._height = self.line_height + t+b
        ctx.restore()


    def render(self, ctx):
        ctx.save()

        ctx.rectangle(*self.bounds)
        ctx.set_source_rgb(0.9, 0.9, 0.8)
        ctx.stroke_preserve()
        ctx.set_source_rgb(1, 1, 0.9)
        ctx.fill()

        ctx.set_source_rgb(0.2, 0.2, 0.2)
        ctx.set_font_size(self.font_size)
        if self.bold:
            ctx.set_font("Helvetica", False, self.bold)
        t, l, b, r = self.padding
        x, y = self.pos
        w, h = self.size
        ctx.show_text_lines_at(int(x+l)+0.5, int(y+t)+0.5, self.text, self.line_spacing, self.line_height)
        ctx.stroke()
        ctx.restore()


    def do_relayout(self, cr):
        self.calc(cr)



# Copyright (c) 2007, 2016, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import weakref

from mforms import newLabel
import mforms

#-------------------------------------------------------------------------------

def weakcb(object, cbname):
    """Create a callback that holds a weak reference to the object. When passing a callback
    for mforms, use this to create a ref to it and prevent circular references that are never freed.
    """
    def call(ref, cbname):
        callback = getattr(ref(), cbname, None)
        if callback is None:
            print "Object has no callback %s"%cbname
        else:
            return callback()

    return lambda ref=weakref.ref(object): call(ref, cbname)


not_running_warning_label_text = "There is no connection to the MySQL Server.\nThis functionality requires an established connection to a running MySQL server."
def not_running_warning_label():
    warning = newLabel("\n\n\n\n"+not_running_warning_label_text)
    warning.set_style(mforms.BigStyle)
    warning.set_text_align(mforms.MiddleCenter)
    warning.show(False)
    return warning

def no_remote_admin_warning_label(server_instance_settings):
    if server_instance_settings.uses_ssh:
        warning = newLabel("There is no SSH connection to the server.\nTo use this functionality, the server where MySQL is located must have a SSH server running\nand you must provide its login information in the server profile.")
    else:
        if server_instance_settings.uses_wmi:
            warning = newLabel("There is no WMI connection to the server.\nTo use this functionality, the server where MySQL is located must be configured to use WMI\nand you must provide its login information in the server profile.")
        else:
            warning = newLabel("Remote Administration is disabled.\nTo use this functionality, the server where MySQL is located must either have an SSH server running,\nor if it is a Windows machine, must have WMI enabled.\nAdditionally you must enable remote administration in the server profile, providing login details for it.")
    warning.set_style(mforms.BigStyle)
    warning.set_text_align(mforms.MiddleCenter)
    return warning


def make_panel_header(icon, title, subtitle, button=None):
    table = mforms.newTable()
    table.set_row_count(2)
    table.set_column_count(3)
    table.set_row_spacing(0)
    table.set_column_spacing(15)
    image = mforms.newImageBox()
    image.set_image(mforms.App.get().get_resource_path(icon))
    image.set_image_align(mforms.TopCenter)
    table.add(image, 0, 1, 0, 2, mforms.HFillFlag)
    label = mforms.newLabel(title)
    label.set_text_align(mforms.BottomLeft)
    label.set_style(mforms.SmallStyle)
    table.add(label, 1, 2, 0, 1, mforms.HFillFlag|mforms.HExpandFlag)
    label = mforms.newLabel(subtitle)
    label.set_text_align(mforms.TopLeft)
    label.set_style(mforms.VeryBigStyle)
    table.add(label, 1, 2, 1, 2, mforms.HFillFlag|mforms.HExpandFlag)
    if button:
        table.add(button, 2, 3, 0, 2, mforms.HFillFlag)
    return table

def show_error_page(parent, text):

    if not hasattr(parent, '_error_label'):
        parent._error_label = None

    if parent._error_label is None:
        parent._error_label = mforms.newLabel(text)
        parent._error_label.set_style(mforms.BoldStyle)
        parent._error_label.set_text_align(mforms.MiddleCenter)
        parent.add(parent._error_label, True, True)

def remove_error_page_if_exists(parent):

    if not hasattr(parent, '_error_label'):
        parent._error_label = None

    if parent._error_label:
        parent.remove(parent._error_label)



class WbAdminBaseTab(mforms.Box):
    ui_created = False
    warning_panel = None
    content = None

    def __init__(self, ctrl_be, instance_info, main_view):
        mforms.Box.__init__(self, False)
        self.set_managed()
        self.set_release_on_add()
        
        self.set_spacing(12)
        
        self.instance_info = instance_info
        self.ctrl_be = ctrl_be
        self.main_view = main_view
    
    
    @property
    def editor(self):
        return self.main_view.editor
    

    def create_basic_ui(self, icon, title, button=None):
        self.heading = make_panel_header(icon, self.instance_info.name, title, button)
        self.heading.set_padding(12)
        self.add(self.heading, False, False)
        
        self.content = None
    
    
    def page_activated(self):
        if self.warning_panel:
            self.remove(self.warning_panel)
            self.warning_panel = None

        button_data = None
        self.page_active = True
        if not self.ctrl_be.is_sql_connected():
            if self.content:
                self.content.show(False)
            text = ("There is no connection to the MySQL Server.", "This functionality requires an established connection to a running MySQL server to work.")
        else:
            text = None
          
        if text:
            title, text = text
            self.warning_panel = MessageButtonPanel("", title, text, button_data)
            self.add(self.warning_panel, True, True)
        else:
            if not self.ui_created:
              self.suspend_layout()
              self.create_ui()
              self.resume_layout()
              self.ui_created = True

    def get_select_int_result(self, query, column=0):
        res = self.main_view.editor.executeManagementQuery(query, 0)
        if res and res.goToFirstRow():
            return res.intFieldValue(column)
        return None


class MessageButtonPanel(mforms.Table):
    def __init__(self, icon, title, text, button1, button2=None):
        mforms.Table.__init__(self)
        self.set_managed()
        self.set_release_on_add()
        
        self.set_padding(-1) # center contents
        self.set_column_spacing(12)
        self.set_row_spacing(12)
        self.set_row_count(3)
        self.set_column_count(2)
        
        if icon:
            image = mforms.newImageBox()
            image.set_image(mforms.App.get().get_resource_path(icon))
            self.add(image, 0, 1, 0, 3, 0)
        
        
        self._label = mforms.newLabel(title)
        self._label.set_style(mforms.BigBoldStyle)
        self._label.set_text_align(mforms.MiddleCenter)
        self.add(self._label, 1, 2, 0, 1, mforms.HFillFlag)
        
        self.add(mforms.newLabel(text), 1, 2, 1, 2, mforms.HFillFlag)

        if button1 or button2:
            bbox = mforms.newBox(True)
            bbox.set_spacing(12)

        if button1:
            button_caption, button_action = button1
            self._button = mforms.newButton()
            self._button.set_text(button_caption)
            self._button.add_clicked_callback(button_action)
            bbox.add_end(self._button, False, True)
        else:
            self._button = None

        if button2:
            button_caption, button_action = button2
            self._alt_button = mforms.newButton()
            self._alt_button.set_text(button_caption)
            self._alt_button.add_clicked_callback(button_action)
            bbox.add_end(self._alt_button, False, True)
        else:
            self._alt_button = None

        if button1 or button2:
            self.add(bbox, 1, 2, 2, 3, 0)



# Copyright (c) 2014 Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import os

import re
import grt
import mforms

from threading import Thread
from Queue import Queue, Empty

from workbench.log import log_info, log_error
from workbench.client_utils import MySQLScriptImporter


class RunPanel(mforms.Table):
    def __init__(self, editor, log_callback):
        mforms.Table.__init__(self)
        self.set_managed()
        self.set_release_on_add()

        self.set_row_count(2)
        self.set_column_count(1)

        self.set_padding(-1)

        self.label = mforms.newLabel("Running script...")
        self.add(self.label, 0, 1, 0, 1, mforms.HFillFlag)
        
        self.progress = mforms.newProgressBar()
        self.add(self.progress, 0, 1, 1, 2, mforms.HFillFlag)
        
        self.progress.set_size(400, -1)
        
        self.log_callback = log_callback
        self.editor = editor
        
        self.importer = MySQLScriptImporter(editor.connection)
        self.importer.report_progress = self.report_progress
        self.importer.report_output = self.report_output
        
        self._worker_queue = Queue()
        self._worker = None
        self._progress_status = None
        self._progress_value = 0

        self._update_timer = None

    def __del__(self):
        if self._update_timer:
            mforms.Utilities.cancel_timeout(self._update_timer)

    @property
    def is_busy(self):
        return self._worker != None


    def report_progress(self, status, current, total):
        self._progress_status = status
        if status:
            self._worker_queue.put(status)
        if total > 0:
            self._progress_value = float(current) / total


    def report_output(self, message):
        log_info("%s\n" % message)
        self._worker_queue.put(message)


    def update_ui(self):
        try:
            while True:
                data = self._worker_queue.get_nowait()
                if data is None:
                    self._worker.join()
                    self._worker = None
                    self._update_timer = None
                    self.progress.show(False)
                    self.log_callback(None)
                    return False

                if isinstance(data, Exception):
                    self.log_callback(str(data)+"\n")
                    if isinstance(data, grt.DBError) and data.args[1] == 1044:
                        mforms.Utilities.show_error("Run SQL Script",
                                                    "The current MySQL account does not have enough privileges to execute the script.", "OK", "", "")
                    elif isinstance(data, grt.DBLoginError):
                        username = self.editor.connection.parameterValues["userName"]
                        host = self.editor.connection.hostIdentifier
                        mforms.Utilities.forget_password(host, username)

                        mforms.Utilities.show_error("Run SQL Script",
                                                    "Error executing SQL script.\n"+str(data), "OK", "", "")
                    else:
                        mforms.Utilities.show_error("Run SQL Script",
                                                    "Error executing SQL script.\n"+str(data), "OK", "", "")
                else:
                    self.log_callback(data+"\n")

                if self._progress_status is not None:
                    self.label.set_text(self._progress_status)
                self.progress.set_value(self._progress_value)
        except Empty:
            pass

        if self._progress_status is not None:
            self.label.set_text(self._progress_status)
        self.progress.set_value(self._progress_value)

        return True
            
            
    def work(self, file, schema, charset):
        try:
            log_info("Executing %s...\n" % file)
            self._progress_status = "Executing %s..." % file
            self._progress_value = 0
            self.importer.import_script(file, default_schema = schema, default_charset = charset)

            log_info("Run script finished\n")
        except grt.DBLoginError, e:
            log_error("MySQL login error running script: %s\n" % e)
            self._worker_queue.put(e)
        except grt.DBError, e:
            log_error("MySQL error running script: %s\n" % e)
            self._worker_queue.put(e)
        except Exception, e:
            import traceback
            log_error("Unexpected exception running script: %s\n%s\n" % (e, traceback.format_exc()))
            self._worker_queue.put(e)
        self._worker_queue.put(None)


    def start(self, what, default_db, default_charset):
        parameterValues = self.editor.connection.parameterValues
        pwd = parameterValues["password"]
        if not pwd:
            username = parameterValues["userName"]
            host = self.editor.connection.hostIdentifier

            ok, pwd = mforms.Utilities.find_cached_password(host, username)
            if not ok:
                accepted, pwd = mforms.Utilities.find_or_ask_for_password("Run SQL Script", host, username, False)
                if not accepted:
                    return
        self.importer.set_password(pwd)

        self._worker = Thread(target = self.work, args = (what, default_db, default_charset))
        self._worker.start()
    
        self._update_timer = mforms.Utilities.add_timeout(0.2, self.update_ui)




class ParameterDialog(mforms.Form):
    def __init__(self, editor):
        mforms.Form.__init__(self, mforms.Form.main_form(), mforms.FormDialogFrame)

        self.editor = editor
        box = mforms.Box(False)
        box.set_padding(12)
        box.set_spacing(8)

        box.add(mforms.newLabel("Preview the first lines of the script below and click [Run] to start executing.\nNote: the preview below may display non-ASCII characters incorrectly, even if the MySQL server can treat them correctly."), False, True)
        self.file_info = mforms.newLabel("")
        box.add(self.file_info, False, True)
        self.text = mforms.newCodeEditor(None)
        self.text.set_language(mforms.LanguageMySQL)
        
        box.add(self.text, True, True)


        table = mforms.newTable()
        table.set_padding(20)
        table.set_row_count(2)
        table.set_column_count(3)
        table.set_row_spacing(8)
        table.set_column_spacing(4)
        table.add(mforms.newLabel("Default Schema Name:"), 0, 1, 0, 1, 0)
        self.schema = mforms.newSelector(mforms.SelectorCombobox)
        table.add(self.schema, 1, 2, 0, 1, mforms.HFillFlag|mforms.HExpandFlag)

        help = mforms.newLabel("Schema to be used unless explicitly specified in the script.\nLeave blank if the script already specified it,\npick a schema from the drop down or type a name to\ncreate a new one.")
        help.set_style(mforms.SmallHelpTextStyle)
        table.add(help, 2, 3, 0, 1, mforms.HFillFlag)

        table.add(mforms.newLabel("Default Character Set:"), 0, 1, 1, 2, 0)
        self.charset = mforms.newSelector()
        self.charset.add_changed_callback(self.update_preview)
        l = [""]
        for ch in grt.root.wb.rdbmsMgmt.rdbms[0].characterSets:
            l.append(ch.name)
        self.charset.add_items(sorted(l))
        table.add(self.charset, 1, 2, 1, 2, mforms.HFillFlag|mforms.HExpandFlag)

        help = mforms.newLabel("Default character set to use when executing the script,\nunless specified in the script.")
        help.set_style(mforms.SmallHelpTextStyle)
        table.add(help, 2, 3, 1, 2, mforms.HFillFlag)

        box.add(table, False, True)

        self.ok = mforms.newButton()
        self.ok.set_text("Run")

        self.cancel = mforms.newButton()
        self.cancel.set_text("Cancel")

        hbox = mforms.Box(True)
        hbox.set_spacing(8)
        mforms.Utilities.add_end_ok_cancel_buttons(hbox, self.ok, self.cancel)
        box.add_end(hbox, False, True)

        self.set_content(box)

    def run(self, file):
        self.set_title("Run SQL Script - %s" % file)

        known_schemas = [""]
        result = self.editor.executeManagementQuery("SHOW SCHEMAS", 0)
        if result:
            while result.nextRow():
                s = result.stringFieldValue(0)
                if s not in ["performance_schema", "mysql", "information_schema"]:
                    known_schemas.append(s)
            self.schema.add_items(sorted(known_schemas))

        try:
            self.preview_data = open(file).read(4098)
            self.detected_charset = None
            r = re.compile("/\*[^ ]* SET NAMES ([a-zA-Z0-9_]*) \*/")
            # try to detect the charset name from the SET NAMES line
            for line in self.preview_data.split("\n"):
                if line.startswith("/*"):
                    g = r.match(line)
                    if g:
                        c = g.groups()[0]
                        log_info("Character set of file %s detected to be %s\n" % (file, c))
                        self.detected_charset = c
                        break

            self.file_info.set_text("%i total bytes in file, displaying first %i bytes" % (os.stat(file).st_size, len(self.preview_data)))
            self.update_preview()
        except Exception, e:
            mforms.Utilities.show_error("Run SQL Script", str(e), "OK", "", "")
            return

        self.set_size(800, 600)
        self.center()
        return self.run_modal(self.ok, self.cancel)


    def get_default_charset(self):
        c = self.charset.get_string_value().encode("utf8")
        if c == "Default":
            return ""
        return c

    def get_default_schema(self):
        return self.schema.get_string_value().encode("utf8")

    def update_preview(self):
        c = self.charset.get_string_value()
        data = self.preview_data
        if c and c != "utf8":
            tmp = data
            while tmp:
                try:
                    data = tmp.decode(c).encode("utf8")
                    break
                except (UnicodeDecodeError, LookupError):
                    tmp = tmp[:-1]
            if not tmp:
                log_error("Could not convert file %s from %s to utf8\n" % (file, c))
        self.text.set_features(mforms.FeatureReadOnly, False)
        self.text.set_text(data)
        self.text.set_features(mforms.FeatureReadOnly, True)


class RunScriptForm(mforms.Form):
    def __init__(self, editor):
        mforms.Form.__init__(self, mforms.Form.main_form(), mforms.FormDialogFrame)

        self.editor = editor
        self.logbox = mforms.newTextBox(mforms.VerticalScrollBar)


    def report(self, text):
        if text is None:
            self.ok.set_enabled(True)
        else:
            self.logbox.append_text_and_scroll(text, True)


    def start_import(self, file, default_schema, default_charset):
        self.panel = RunPanel(self.editor, self.report)
        self.set_title("Run SQL Script")

        box = mforms.newBox(False)
        box.set_padding(12)
        box.set_spacing(12)

        box.add(self.panel, False, True)
        box.add(mforms.newLabel("Output:"), False, True)
        box.add(self.logbox, True, True)

        self.ok = mforms.newButton()
        self.ok.set_text("Close")
        self.ok.add_clicked_callback(self.close)

        #self.abort = mforms.newButton()
        #self.abort.set_text("Abort")

        hbox = mforms.Box(True)
        hbox.set_spacing(8)
        #hbox.add(self.abort, False, True)
        hbox.add_end(self.ok, False, True)
        box.add_end(hbox, False, True)

        self.set_content(box)

        self.set_size(800, 600)
        self.center()
        self.show()

        self.ok.set_enabled(False)
        self.panel.start(file, default_schema, default_charset)



    def run(self):
        chooser = mforms.FileChooser(mforms.OpenFile)
        chooser.set_title("Run SQL Script")
        chooser.set_extensions('SQL Scripts (*.sql)|*.sql', 'sql')
        if chooser.run_modal():
            dlg = ParameterDialog(self.editor)
            if dlg.run(chooser.get_path()):
                schema = dlg.get_default_schema()
                if schema:
                    self.editor.executeManagementCommand("CREATE SCHEMA IF NOT EXISTS `%s`" % schema, 1)

                self.start_import(chooser.get_path().encode("utf8"), dlg.get_default_schema(), dlg.get_default_charset())
                return True
        return False


    def run_file(self, path):
        dlg = ParameterDialog(self.editor)
        if dlg.run(path):
            schema = dlg.get_default_schema()
            if schema:
                self.editor.executeManagementCommand("CREATE SCHEMA IF NOT EXISTS `%s`" % schema, 1)

            self.start_import(path, dlg.get_default_schema(), dlg.get_default_charset())
            return True
        return False




# Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA


import re
import grt
import mforms
from workbench.db_utils import escape_sql_identifier
from workbench.plugins import insert_item_to_plugin_context_menu

from table_templates import TableTemplateManager

from sqlide_tableman_ext import CreateIndexForm
from sqlide_catalogman_ext import show_schema_manager

def esc_ident(s):
    # escape SQL identifier and add backtics only if needed
    if not re.match("^([a-z][A-Z][0-9]_)*$", s):
        return "`%s`" % escape_sql_identifier(s)
    return s


def _(s):
    return s

template_manager = None
def init():
    global template_manager

    template_manager = TableTemplateManager()


def handleLiveTreeContextMenu(name, sender, args):
    menu = mforms.fromgrt(args['menu'])
    selection = args['selection']

    # Add extra menu items to the SQL editor live schema tree context menu

    node_type = None
    mixed = False
    selection_type = None
    object_selected = False
    object_selected_count = 0
    column_selected = False
    if len(selection) == 1:
        object_type_caption = {
            'db.Schema' : "Schema",
            'db.Table' : "Table",
            'db.View' : "View",
            'db.Function' : "Function",
            'db.StoredProcedure' : "Stored Procedure",
            'db.Index' : "Index",
            'db.Trigger' : "Trigger",
            'db.ForeignKey' : "Foreign Key",
        }
    else:
        object_type_caption = {
            'db.Schema' : "Schemas",
            'db.Table' : "Tables",
            'db.View' : "Views",
            'db.Function' : "Functions",
            'db.StoredProcedure' : "Stored Procedures",
            'db.Index' : "Indexes",
            'db.Trigger' : "Triggers",
            'db.ForeignKey' : "Foreign Keys",
        }
    unique_tables = set()
    selection_db_type = None
    for s in selection:
        if s.type != node_type:
            if node_type is None:
                node_type = s.type
            else:
                mixed = True
        if s.type == 'columns':
            selection_type = s.type
            column_selected = True
            unique_tables.add(s.name)
        elif s.type in ['tables', 'views', 'functions', 'storedProcedures']:
            selection_type = s.type
            object_selected = False
            if s.type == 'tables':
                selection_db_type = 'db.Table'
            elif s.type == 'views':
                selection_db_type = 'db.View'
            elif s.type == 'functions':
                selection_db_type = 'db.Function'
            elif s.type == 'storedProcedures':
                selection_db_type = 'db.StoredProcedure'
        elif s.type in ('db.Schema', 'db.Table', 'db.View', 'db.Function', 'db.StoredProcedure'):
            selection_type = s.type
            selection_db_type = s.type
            object_selected = True
            object_selected_count += 1
            if s.type == 'db.Table':
                unique_tables.add(s.name)
        elif s.type in ('db.Index', 'db.Trigger'):
            selection_type = s.type
            object_selected = False
        elif s.type == ('db.Column'):
            selection_type = s.type
            if s.owner:
                unique_tables.add(s.owner.name)
                selection_type = s.owner.type+':'+s.type
            column_selected = True
        else:
            print "Unhandled type", s.type

    if mixed:
        selection_type = None
        pass
    else:
        item = menu.find_item('builtins_separator')
        if item:
            index = menu.get_item_index(item)+1
        else:
            index = 0
        needs_separator = False

        if False and object_selected and selection_type == 'db.Table':
            item = mforms.newMenuItem(_("Export Data to CSV File..."))
            item.set_name("export_table_csv")
            menu.insert_item(index-1, item)
            index += 1
            item = mforms.newMenuItem(_("Import Data from CSV File..."))
            item.set_name("import_table_csv")
            menu.insert_item(index-1, item)
            index += 1
            item = mforms.newMenuItem(_("Dump Data to SQL File..."))
            item.set_name("dump_table")
            menu.insert_item(index-1, item)
            index += 1

        if object_selected or column_selected:
            # Add menu items for generating code
            code_items = [
                          # Caption, callback name, supported node types, enable condition, mixed node types allowed
                          (_("Name"), 'name_short', ['db.Schema'], len(selection) > 0, False),
                          (_("Name (short)"), 'name_short', ['tables', 'views', 'functions', 'storedProcedures', 'db.Table', 'db.Table:db.Column', 'db.View:db.Column', 'db.View', 'db.StoredProcedure', 'db.Function'], len(selection) > 0, True),
                          (_("Name (long)"), 'name_long', ['tables', 'views', 'functions', 'storedProcedures', 'db.Table', 'db.Table:db.Column', 'db.View:db.Column', 'db.View', 'db.StoredProcedure', 'db.Function'], len(selection) > 0, True),
                          (_("Select All Statement"), 'select_all_statement', ['db.Table', 'db.View'], len(selection) > 0, False),
                          (_("Select All Statement"), 'select_all_statement', ['columns'], len(selection) == 1 and selection[0].type == 'columns', False),
                          (_("Select Columns Statement"), 'select_columns_statement', ['db.Table:db.Column', 'db.View:db.Column'], len(unique_tables) == 1, False),
                          (_("Insert Statement"), 'insert_all_statement', ['db.Table', 'columns'], len(selection) > 0, False),
                          (_("Insert Statement"), 'insert_columns_statement', ['db.Table:db.Column'], len(selection) > 0, False),
                          (_("Update Statement"), 'update_all_statement', ['db.Table', 'columns'], len(selection) > 0, False),
                          (_("Update Statement"), 'update_columns_statement', ['db.Table:db.Column'], len(selection) > 0, False),
                          (_("Delete Statement"), 'delete_statement', ['db.Table'], len(selection) > 0, False),
                          (_("Create Statement"), 'create_statement', ['db.Schema', 'db.Table', 'db.View', 'db.StoredProcedure', 'db.Function'], len(selection) > 0, False),
                          (_("Procedure Call"), 'call_procedure', ['db.StoredProcedure'], len(selection) > 0, False),
                          (_("Function Call"), 'call_function', ['db.Function'], len(selection) > 0, False),
                          (None, None, ['db.Table'], None, False),
                          (_("Join Selected Tables"), 'build_joined_select', ['db.Table'], len(selection) == 2, False),
                          (_("Delete with References"), 'build_cascaded_delete', ['db.Table'], len(selection) == 1, False),
                          (_("Select Row References"), 'build_cascaded_select', ['db.Table'], len(selection) == 1, False),
                          ]

            copy_submenu = mforms.newMenuItem("Copy to Clipboard")
            copy_submenu.set_name("copy_to_clipboard")
            menu.insert_item(index, copy_submenu)
            index += 1

            send_submenu = mforms.newMenuItem("Send to SQL Editor")
            send_submenu.set_name("send_to_editor")
            menu.insert_item(index, send_submenu)
            index += 1

            gencopy = CodeGenerator(sender, selection, False)
            gensend = CodeGenerator(sender, selection, True)

            for caption, name, types, enabled, allow_mixed in code_items:
                if not allow_mixed and mixed:
                    continue
                if types and selection_type not in types:
                    continue
                if caption is None:
                    copy_submenu.add_separator()
                    send_submenu.add_separator()
                else:
                    item = copy_submenu.add_item_with_title(caption, getattr(gencopy, name), name)
                    item.set_enabled(bool(enabled))

                    item = send_submenu.add_item_with_title(caption, getattr(gensend, name), name)
                    item.set_enabled(bool(enabled))
            needs_separator = True
        else:
            needs_separator = False

        if selection_type in ('tables', 'views', 'functions', 'storedProcedures', 'db.Schema', 'db.Table', 'db.View', 'db.Function', 'db.StoredProcedure') and len(selection) == 1:
            if needs_separator:
                menu.insert_item(index, mforms.newMenuItem("", mforms.SeparatorMenuItem))
                index += 1
                needs_separator = False

            item = mforms.newMenuItem("Create %s..." % object_type_caption.get(selection_db_type, selection_db_type))
            item.add_clicked_callback(lambda: do_create_object(sender, selection[0].schemaName, selection_db_type))
            menu.insert_item(index, item)
            index+= 1

            if selection_type in ('db.Table', 'tables'):
                item = mforms.newMenuItem("Create Table Like...")
                menu.insert_item(index, item)
                index += 1
                #        it = item.add_item_with_title("Selected Table", lambda: template_manager.create_table_like(sender, selection[0].schemaName, selection[0].name))
                #        if selection[0].type == 'tables':
                #            it.set_enabled(False)
                for templ in template_manager.templates:
                    item.add_item_with_title(templ.name, lambda templ=templ, schema=selection[0].schemaName: template_manager.create_table_like_template(sender, schema, templ))
                item.add_separator()
                item.add_item_with_title("Edit Templates...", template_manager.edit_templates)
                
        if selection_type is None:
            if needs_separator:
                menu.insert_item(index, mforms.newMenuItem("", mforms.SeparatorMenuItem))
                index += 1
                needs_separator = False

            item = mforms.newMenuItem("Create Schema...")
            item.add_clicked_callback(lambda: do_create_object(sender, None, "db.Schema"))
            menu.insert_item(index, item)
            index+= 1

        if object_selected:
            if needs_separator:
                menu.insert_item(index, mforms.newMenuItem("", mforms.SeparatorMenuItem))
                index += 1
                needs_separator = False

            if object_selected_count == 1:
                item = mforms.newMenuItem("Alter %s..." % object_type_caption.get(selection_type, selection_type))
            else:
                item = mforms.newMenuItem("Alter %i %s..." % (object_selected_count, object_type_caption.get(selection_type, selection_type)))
            item.add_clicked_callback(lambda: do_alter_object(sender, selection))
            menu.insert_item(index, item)
            index+= 1
            needs_separator = True

        if selection_type == 'db.Table' and object_selected:
            item = mforms.newMenuItem("Table Maintenance...")
            item.add_clicked_callback(lambda: show_schema_manager(sender, set([obj.schemaName for obj in selection]), True))
            menu.insert_item(index, item)
            index += 1

        if object_selected: # or selection_type in ('db.Index', 'db.Trigger'):  <-- enable this once we support selective refresh of LST on drop of these objects
            if needs_separator:
                menu.insert_item(index, mforms.newMenuItem("", mforms.SeparatorMenuItem))
                index += 1
                needs_separator = False

            if object_selected_count == 1:
                item = mforms.newMenuItem("Drop %s..." % object_type_caption.get(selection_type, selection_type))
            else:
                item = mforms.newMenuItem("Drop %i %s..." % (object_selected_count, object_type_caption.get(selection_type, selection_type)))
            item.add_clicked_callback(lambda: do_drop_object(sender, selection))
            menu.insert_item(index, item)
            index+= 1

        if selection_type == 'db.Table' and object_selected:
            item = mforms.newMenuItem("Truncate %s..." % ("Table" if object_selected_count == 1 else "%i Tables" % object_selected_count))
            item.add_clicked_callback(lambda: do_truncate_table(sender, selection))
            menu.insert_item(index, item)
            index += 1

        if selection_type in ('db.Table', 'db.Schema', 'tables'):
            item = mforms.newMenuItem("Search Table Data...")
            item.add_clicked_callback(lambda: open_search(sender))
            insert_item_to_plugin_context_menu(menu, item)

        if selection_type == 'db.Table:db.Column' and column_selected:
            if needs_separator:
                menu.insert_item(index, mforms.newMenuItem("", mforms.SeparatorMenuItem))
                index += 1
                needs_separator = False
            item = mforms.newMenuItem("Create Index...")
            item.add_clicked_callback(lambda: do_create_index(sender, selection))
            menu.insert_item(index, item)
            index += 1


def open_search(editor):
    grt.modules.MySQLDBSearchModule.showSearchPanel(editor)


def do_create_object(editor, schema_name, db_type):
    if db_type == 'db.Schema':
        editor.alterLiveObject(db_type, "", "")
    else:
        editor.alterLiveObject(db_type, schema_name, "")

def do_alter_object(editor, selection):
    for obj in selection:
        editor.alterLiveObject(obj.type, obj.schemaName, obj.name)



def review_sql_code(code):
    form = mforms.Form(mforms.Form.main_form(), mforms.FormNormal)
    form.set_title("Review SQL Code to Execute")
    box = mforms.newBox(False)
    box.set_padding(12)
    form.set_content(box)

    box.set_spacing(8)
    box.add(mforms.newLabel("Review the SQL code to be executed."), False, True)

    editor = mforms.CodeEditor()
    box.add(editor, True, True)
    editor.set_language(mforms.LanguageMySQL)
    editor.set_text(code)
    editor.set_features(mforms.FeatureReadOnly, True)

    ok = mforms.newButton()
    ok.set_text("Execute")
    cancel = mforms.newButton()
    cancel.set_text("Cancel")

    bbox = mforms.newBox(True)
    bbox.set_spacing(8)
    bbox.add_end(ok, False, True)
    bbox.add_end(cancel, False, True)
    box.add_end(bbox, False, True)

    form.set_size(500, 360)
    return form.run_modal(ok, cancel)



def do_truncate_table(editor, selection):
    tables = []
    for obj in selection:
        if obj.type == 'db.Table':
            tables.append("%s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name)))

    if len(tables) == 1:
        res = mforms.Utilities.show_message("Truncate Table", "Please confirm permanent deletion of all rows from table `%s`." % tables[0],
                                         "Review SQL", "Cancel", "Truncate")
    elif len(tables) > 1:
        res = mforms.Utilities.show_message("Truncate Tables", "Please confirm permanent deletion of all rows from %s selected tables." % len(tables),
                                         "Review SQL", "Cancel", "Truncate")
    else:
        return
    if res == mforms.ResultCancel:
        return
    elif res == mforms.ResultOk:
        if not review_sql_code("\n".join([("TRUNCATE %s;" % t) for t in tables])):
            return

    count = 0
    for table in tables:
        try:
            stmt = "TRUNCATE %s" % table
            editor.executeManagementCommand(stmt, 1)
            count += 1
        except Exception, exc:
            if count < len(tables)-1:
                if mforms.Utilities.show_error("Could not Truncate Table", str(exc)+"\nClick Cancel to stop truncating other tables.\n\n"+stmt, "OK", "Cancel", "") == mforms.ResultCancel:
                    break
            else:
                if mforms.Utilities.show_error("Could not Truncate Table", str(exc)+"\n\n"+stmt, "OK", "", "") == mforms.ResultCancel:
                    break
    mforms.App.get().set_status_text("%i tables truncated" % count)


def do_create_index(editor, selection):
    cols = []
    schema = None
    table = None
    for node in selection:
        if schema and schema != node.schemaName:
            mforms.Utilities.show_error("Create Index", "Please select one or more columns from the same table.", "OK", "", "")
            return
        if table and (not node.owner or table != node.owner.name):
            mforms.Utilities.show_error("Create Index", "Please select one or more columns from the same table.", "OK", "", "")
            return
        schema = node.schemaName
        table = node.owner.name
        cols.append(node.name)
    if cols:
        form = CreateIndexForm(mforms.Form.main_form(), editor, schema, table, cols, None)
        if form.run():
            pass



def do_drop_object(editor, selection):
    statements = []
    table_names = []
    object_types = set()
    for obj in selection:
        if obj.type == 'db.Schema':
            stmt = "DROP DATABASE %s" % esc_ident(obj.name)
            object_types.add("Schema")
        elif obj.type == 'db.Table':
            table_names.append("%s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name)))
            object_types.add("Table")
            continue
        elif obj.type == 'db.Trigger':
            stmt = "DROP TRIGGER %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name))
            object_types.add("Trigger")
        elif obj.type == 'db.StoredProcedure':
            stmt = "DROP PROCEDURE %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name))
            object_types.add("Stored Procedure")
        elif obj.type == 'db.Function':
            stmt = "DROP FUNCTION %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name))
            object_types.add("Function")
        elif obj.type == 'db.View':
            stmt = "DROP VIEW %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name))
            object_types.add("View")
        elif obj.type == 'db.Index':
            stmt = "DROP INDEX %s ON %s.%s" % (esc_ident(obj.name), esc_ident(obj.owner.schemaName), esc_ident(obj.owner.name))
            object_types.add("Index")
        else:
            print "Unsupported type for drop", obj.type
        statements.append((1, stmt))

    # multiple tables can be dropped in the same command (probably so that we dont get ref constraint errors)
    if table_names:
        stmt = "DROP TABLE %s" % ", ".join(table_names)
        count = len(statements) + len(table_names)
        statements.append((len(table_names), stmt))
    else:
        count = len(statements)

    if not count:
        return

    # don't let dropping tables or schemas mixed with other objects to make the confirmation message clearer and reduce risk of accident
    if len(object_types) > 1:
        mforms.Utilities.show_message("Drop Objects", "Multiple objects of different types are selected. To avoid accidents, please drop one type of object at a time.", "OK", "", "")
        return

    res = None
    if "Table" in object_types:
        if count == 1:
            res = mforms.Utilities.show_message("Drop Table",
                                             "Please confirm permanent deletion of table `%s` and its data." % (selection[0].name),
                                             "Review SQL", "Cancel", "Drop Now")
        else:
            res = mforms.Utilities.show_message("Drop %i Tables" % len(table_names),
                                             "Please confirm permanent deletion of tables %s and their data." % ", ".join(table_names),
                                             "Review SQL", "Cancel", "Drop Now")
    elif "Schema" in object_types:
        if count == 1:
            res = mforms.Utilities.show_message("Drop Schema",
                                             "Please confirm permanent deletion of schema `%s` and all its data." % (selection[0].name),
                                             "Review SQL", "Cancel", "Drop Now")
        else:
            res = mforms.Utilities.show_message("Drop %i Schemas" % count,
                                             "Please confirm permanent deletion of schemas %s and all their data." % (", ".join([s.name for s in selection])),
                                             "Review SQL", "Cancel", "Drop Now")
    else:
        if count == 1:
            res = mforms.Utilities.show_message("Drop %s" % list(object_types)[0],
                                             "Please confirm permanent deletion of %s `%s`." % (list(object_types)[0].lower(), selection[0].name),
                                             "Review SQL", "Cancel", "Drop Now")
        else:
            res = mforms.Utilities.show_message("Drop Objects", "Please confirm permanent deletion of %s %s objects: %s" % (count, ", ".join(list(object_types)), ", ".join([s.name for s in selection])),
                                             "Review SQL", "Cancel", "Drop Now")
    if res is None or res == mforms.ResultCancel:
        return

    if res == mforms.ResultOk:
        if not review_sql_code("\n".join([s+";" for c, s in statements])):
            return

    drop_count = 0
    for i, (c, stmt) in enumerate(statements):
        try:
            #This should be run in background because of the bug: #71327
            editor.executeCommand(stmt,1 ,1)
            drop_count += c
        except Exception, exc:
            if i < len(statements)-1:
                if mforms.Utilities.show_error("Could not Drop Object", str(exc)+"\nClick Cancel to stop dropping other objects.\n\n"+stmt, "OK", "Cancel", "") == mforms.ResultCancel:
                    break
            else:
                if mforms.Utilities.show_error("Could not Drop Object", str(exc)+"\n\n"+stmt, "OK", "", "") == mforms.ResultCancel:
                    break
    mforms.App.get().set_status_text("%i objects dropped" % drop_count)


class DependencyAnalyzer:
    def __init__(self, catalog, schema_name):
        self.schema_name = schema_name
        self.catalog = catalog
        self.tables_by_name = {}
        self.referencing_foreign_keys = {}


    def load_data(self, connection):
        m = grt.modules.DbMySQLRE
        grt.log_info("sqlide_grt", "Connecting...")
        ok, password = mforms.Utilities.find_or_ask_for_password("Connect to MySQL", connection.hostIdentifier, "root", False)
        if not ok:
            return None
        m.connect(connection, password)

        grt.log_info("sqlide_grt", "Reverse engineering schema %s...\n" % self.schema_name)
        options = {"reverseEngineerTables" : True,
            "reverseEngineerTriggers" : False,
            "reverseEngineerViews" : False,
            "reverseEngineerRoutines" : False}
        self.catalog = m.reverseEngineer(connection, "", [self.schema_name], options)

        m.disconnect(connection)
        grt.log_info("sqlide_grt", "Reverse engineer schema %s finished\n" % self.schema_name)

        return self.catalog


    def scan_tables(self):
        # build a map of all tables and the FKs that reference them
        for s in self.catalog.schemata:
            for t in s.tables:
                self.tables_by_name[t.name] = t

                for f in t.foreignKeys:
                    refs = self.referencing_foreign_keys.get(f.referencedTable, [])
                    refs.append(f)
                    self.referencing_foreign_keys[f.referencedTable] = refs


    def table_with_name(self, schema, name):
        if schema == self.schema_name:
            return self.tables_by_name.get(name, None)
        return None


    def get_foreign_keys_to_table(self, table):
        return self.referencing_foreign_keys.get(table, [])


    def get_referenced_tables(self, table):
        return [fk.referencedTable for fk in table.foreignKeys]


    def get_foreign_key_between(self, table1, table2):
        for fk in table1.foreignKeys:
            if fk.referencedTable == table2:
                return fk
        for fk in table2.foreignKeys:
            if fk.referencedTable == table1:
                return fk
        return None


    def get_foreign_key_from(self, table1, table2):
        for fk in table1.foreignKeys:
            if fk.referencedTable == table2:
                return fk
        return None


    def find_foreign_key_path_between(self, from_table, to_table):
        # find the shortest path between the 2 given tables using Dijkstra's algorithm
        # table = vertex, FK = edge

        predecessor = {}
        distance = {}

        # initialization
        for table in self.tables_by_name.values():
            predecessor[table] = None
            distance[table] = 99999999
        distance[from_table] = 0

        def relax(u, v):
            if distance[v] > distance[u] + 1:
                distance[v] = distance[u] + 1
                predecessor[v] = u

        def get_cheapest(remaining):
            v = reduce(lambda a, b: a if distance[a] < distance[b] else b, remaining)
            remaining.remove(v)
            return v

        # main loop
        remaining = self.tables_by_name.values()
        while remaining:
            u = get_cheapest(remaining)
            for v in self.get_referenced_tables(u):
                relax(u, v)

        # get the path
        foreign_keys = []
        t = to_table
        while predecessor[t]:
            pred = predecessor[t]
            foreign_keys.append(self.get_foreign_key_from(pred, t))
            t = pred

        return foreign_keys



def dependencyInfoForSchemaInEditor(editor, schema):
    catalog = editor.customData.get("sqlide_grt:Catalog:%s" % schema, None)
    if not catalog:
        mforms.App.get().set_status_text("Reverse engineering schema for %s..." % schema)

        info = DependencyAnalyzer(None, schema)
        catalog = info.load_data(editor.connection)
        if not catalog:
            mforms.App.get().set_status_text("Cancelled.")
            return

        mforms.App.get().set_status_text("%s reverse engineered" % schema)

        editor.customData["sqlide_grt:Catalog:%s" % schema] = catalog
    else:
        mforms.App.get().set_status_text("Reusing reverse engineered schema for %s..." % schema)
        info = DependencyAnalyzer(catalog, schema)
    return info


def format_fk_join(fk):
    joins = []
    for i in range(len(fk.referencedColumns)):
        joins.append("%s.%s = %s.%s" % (esc_ident(fk.referencedTable.name), esc_ident(fk.referencedColumns[i].name), esc_ident(fk.owner.name), esc_ident(fk.columns[i].name)))
    return " AND ".join(joins)


def join_tables(info, tables, auto_add_missing):
    joins = set()
    for i, t1 in enumerate(tables):
        for t2 in tables[i+1:]:
            fk = info.get_foreign_key_between(t1, t2)
            if fk:
                joins.add(fk)
                break
            else:
                if auto_add_missing:
                    fks = info.find_foreign_key_path_between(t1, t2) or info.find_foreign_key_path_between(t2, t1)
                    if not fks:
                        print "Could not find path from %s to %s" % (t1.name, t2.name)
                        return None
                    joins.update(fks)
                else:
                    return None

    if not joins:
        return None

    return "SELECT *\n    FROM %s\n    WHERE %s" % (", ".join(set([esc_ident(t.name) for t in tables] + [esc_ident(fk.owner.name) for fk in joins])), " AND ".join(format_fk_join(fk) for fk in joins))


table_column_cache = {}

def tokenize_argument_list(text):
    count = 1
    token = ""
    tokens = []
    for i in range(1, len(text)):
        if text[i] == '(':
            count += 1
        elif text[i] == ')':
            count -= 1
        if count == 0:
            if token:
                tokens.append(token)
            return tokens
        elif count == 1 and text[i] == ',':
            tokens.append(token)
            token = ""
        else:
            token += text[i]
    return None

class CodeGenerator:
    def __init__(self, editor, selection, to_editor):
        self.editor = editor
        self.selection = selection
        self.to_editor = to_editor


    def get_table_columns(self, schema, table):
        info_key = "%s:%s.%s" % (self.editor.__id__, schema, table)
        if table_column_cache.has_key(info_key):
            return table_column_cache[info_key]

        rs = self.editor.executeManagementQuery("SHOW COLUMNS FROM %s.%s" % (esc_ident(schema), esc_ident(table)), 0)
        if rs:
            columns = []
            ok = rs.goToFirstRow()
            while ok:
                name = rs.stringFieldValueByName("Field")
                is_key = rs.stringFieldValueByName("Key") == "PRI"
                default = rs.stringFieldValueByName("Default")
                columns.append((name, is_key, default))
                ok = rs.nextRow()
            table_column_cache[info_key] = columns
            print columns
            return columns
        else:
            return None

    def copy_to_clipboard(self, text):
        mforms.Utilities.set_clipboard_text(text)

    def send_to_editor(self, text):
        ed = self.editor.activeQueryEditor
        if not ed:
            ed = self.editor.addQueryEditor()
        if ed:
            ed.replaceSelection(text)

    def send(self, text):
        if self.to_editor:
            self.send_to_editor(text)
        else:
            self.copy_to_clipboard(text)


    def name_short(self):
        # assumes only real object nodes or Columns node
        parts = []
        for obj in self.selection:
            parts.append("%s" % esc_ident(obj.name))
        self.send(", ".join(parts))

    def name_long(self):
        # assumes only real object nodes of Columns node
        parts = []
        for obj in self.selection:
            parts.append("%s.%s" % (esc_ident(obj.owner.name) if obj.owner else esc_ident(obj.schemaName), esc_ident(obj.name)))
        self.send(", ".join(parts))

    def select_all_statement(self):
        # assumes only table nodes (or the Columns node of a table)
        parts = []
        for obj in self.selection:
            if obj.type == 'columns':
                obj = obj.owner
            parts.append("SELECT %s\nFROM %s.%s;\n" % (",\n    ".join("%s.%s" % (esc_ident(obj.name), esc_ident(c[0])) for c in self.get_table_columns(obj.schemaName, obj.name)), esc_ident(obj.schemaName), esc_ident(obj.name)))
        self.send("\n".join(parts))

    def select_columns_statement(self):
        # assumes only column nodes or Columns parent node, all of a single unique table
        columns = []
        table = None
        for obj in self.selection:
            if obj.type == 'db.Column':
                columns.append(obj.name)
                if obj.owner:
                    table = obj.owner
        if table:
            self.send("SELECT %s FROM %s.%s;\n" % (", ".join("%s" % esc_ident(c) for c in columns), esc_ident(table.schemaName), esc_ident(table.name)))

    def insert_all_statement(self):
        # assumes only table nodes (or the Columns node of a table)
        parts = []
        for obj in self.selection:
            if obj.type == 'columns':
                obj = obj.owner
            columns = self.get_table_columns(obj.schemaName, obj.name)
            parts.append("INSERT INTO %s.%s\n(%s)\nVALUES\n(%s);\n" % (esc_ident(obj.schemaName), esc_ident(obj.name),
                                                                             ",\n".join("%s" % esc_ident(c[0]) for c in columns),
                                                                             ",\n".join("<{%s: %s}>" % (c[0], c[2]) for c in columns)))
        self.send("\n".join(parts))

    def insert_columns_statement(self):
        # assumes only table column nodes, from the same table
        table = self.selection[0].owner
        self.send("INSERT INTO %s.%s\n(%s)\nVALUES\n(%s);\n" % (esc_ident(table.schemaName), esc_ident(table.name),
                                                                    ",\n".join("%s" % esc_ident(obj.name) for obj in self.selection),
                                                                    ",\n".join("<{%s}>" % c.name for c in self.selection)))

    def update_all_statement(self):
        # assumes only table nodes (or the Columns node of a table)
        parts = []
        for obj in self.selection:
            if obj.type == 'columns':
                obj = obj.owner
            columns = self.get_table_columns(obj.schemaName, obj.name)
            parts.append("UPDATE %s.%s\nSET\n%s\nWHERE %s;\n" % (esc_ident(obj.schemaName), esc_ident(obj.name),
                                                                  ",\n".join("%s = <{%s: %s}>" % (esc_ident(c[0]), c[0], c[2]) for c in columns),
                                                                  " AND ".join(["%s = <{expr}>" % esc_ident(c[0]) for c in columns if c[1]])))
        self.send("\n".join(parts))

    def update_columns_statement(self):
        # assumes only table nodes (or the Columns node of a table)
        table = self.selection[0].owner
        self.send("UPDATE %s.%s\nSET\n%s\nWHERE <{where_expression}>;\n" % (esc_ident(table.schemaName), esc_ident(table.name),
                                                                    ",\n".join("%s = <{%s}>" % (esc_ident(obj.name), obj.name) for obj in self.selection)))

    def delete_statement(self):
        parts = []
        for table in self.selection:
            parts.append("DELETE FROM %s.%s\nWHERE <{where_expression}>;\n" % (esc_ident(table.schemaName), esc_ident(table.name)))
        self.send("\n".join(parts))

    def create_statement(self):
        parts = []
        for obj in self.selection:
            wrapper = "%s;\n"
            if obj.type == 'db.Schema':
                rs = self.editor.executeManagementQuery("SHOW CREATE SCHEMA %s" % (esc_ident(obj.name)), 0)
                field_name = 'Create Database'
            elif obj.type == 'db.Table':
                rs = self.editor.executeManagementQuery("SHOW CREATE TABLE %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name)), 0)
                field_name = 'Create Table'
            elif obj.type == 'db.Trigger':
                rs = self.editor.executeManagementQuery("SHOW CREATE TRIGGER %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name)), 0)
                field_name = 'Statement'
                wrapper = "DELIMITER $$\n%s$$\nDELIMITER ;\n"
            elif obj.type == 'db.StoredProcedure':
                rs = self.editor.executeManagementQuery("SHOW CREATE PROCEDURE %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name)), 0)
                field_name = 'Create Procedure'
                wrapper = "DELIMITER $$\n%s$$\nDELIMITER ;\n"
            elif obj.type == 'db.Function':
                rs = self.editor.executeManagementQuery("SHOW CREATE FUNCTION %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name)), 0)
                field_name = 'Create Function'
                wrapper = "DELIMITER $$\n%s$$\nDELIMITER ;\n"
            elif obj.type == 'db.View':
                rs = self.editor.executeManagementQuery("SHOW CREATE VIEW %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name)), 0)
                field_name = 'Create View'
            else:
                print "Unsupported type", obj.type
                continue
            if rs and rs.goToFirstRow():
                sql = rs.stringFieldValueByName(field_name)
                if not sql:
                    print "No field %s for %s" % (field_name, obj.name)
                else:
                    parts.append(wrapper % sql)
            else:
                print "Couldn't fetch create code for %s" % obj.name

        self.send("\n".join(parts))


    def call_procedure(self):
        parts = []
        for obj in self.selection:
            rs = self.editor.executeManagementQuery("SHOW CREATE PROCEDURE %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name)), 0)
            if rs and rs.goToFirstRow():
                sql = rs.stringFieldValueByName('Create Procedure')
                if sql:
                    s = re.match("(CREATE .* PROCEDURE [^(]*\()", sql)
                    if s:
                        args = tokenize_argument_list(sql[len(s.groups()[0])-1:])
                        parts.append("CALL %s.%s(%s);\n" % (esc_ident(obj.schemaName), esc_ident(obj.name),
                                                                ", ".join(["<{%s}>" % a.strip() for a in args])))
        self.send("".join(parts))

    def call_function(self):
        parts = []
        for obj in self.selection:
            rs = self.editor.executeManagementQuery("SHOW CREATE FUNCTION %s.%s" % (esc_ident(obj.schemaName), esc_ident(obj.name)), 0)
            if rs and rs.goToFirstRow():
                sql = rs.stringFieldValueByName('Create Function')
                if sql:
                    s = re.match("(CREATE .* FUNCTION [^(]*\()", sql)
                    if s:
                        args = tokenize_argument_list(sql[len(s.groups()[0])-1:])
                        parts.append("%s.%s(%s)" % (esc_ident(obj.schemaName), esc_ident(obj.name),
                                                        ", ".join(["<{%s}>" % a.strip() for a in args])))
        self.send("\n".join(parts))


    def build_joined_select(self):
        editor = self.editor
        selection = self.selection

        info = dependencyInfoForSchemaInEditor(editor, selection[0].schemaName)
        if not info:
            return
        info.scan_tables()

        tables = []
        for object in selection:
            table = info.table_with_name(object.schemaName, object.name)
            if table:
                tables.append(table)

        if len(tables) != len(selection):
            mforms.Utilities.show_error("Join Selected Tables", "Could reverse engineer some of the selected tables.", "OK", "", "")
            return

        query = join_tables(info, tables, False)
        if query:
            self.send(query)
        else:
            if mforms.Utilities.show_error("Join Selected Tables", "Some of the tables could not be joined together.\nShould missing intermediate tables be searched and added?", "Add Missing", "Cancel", "") == mforms.ResultOk:
                query = join_tables(info, tables, True)
                if query:
                    self.send(query)
                else:
                    mforms.Utilities.show_error("Join Selected Tables", "Could not join selected tables.", "OK", "", "")
                    return
            else:
                return

        if not self.to_editor:
          mforms.App.get().set_status_text("Select statement for joined selection was copied to clipboard")

    def build_cascaded_select(self):
        editor = self.editor
        object = self.selection[0]

        info = dependencyInfoForSchemaInEditor(editor, object.schemaName)
        if not info:
            return
        info.scan_tables()
        the_table = info.table_with_name(object.schemaName, object.name)

        if not the_table:
            mforms.Utilities.show_error("Select Row References", "Could not find reverse engineered %s" % object.name, "OK", "", "")
            return

        if not the_table.primaryKey:
            mforms.Utilities.show_error("Select Row References", "Unable to find referencing tables for %s because the table has no primary key." % object.name,
                                      "OK", "", "")
            return

        # expression to match the row
        pk = []
        for c in the_table.primaryKey.columns:
            pk.append("%s.%s = @%s_to_select" % (the_table.name, c.referencedColumn.name, c.referencedColumn.name))
        pk = " AND ".join(pk)


        def create_select(table, info, selects, reference_chain, visited):
            if table in visited:
                return
            visited.add(table)

            def add_to_list(list, item):
                if item not in list:
                    return list + [item]
                return list

            for fk in info.get_foreign_keys_to_table(table):
                create_select(fk.owner, info, selects, reference_chain + [fk], visited)

            query = "SELECT %s.*\n" % table.name
            query += "    FROM %s\n" % ", ".join(add_to_list([fk.owner.name for fk in reference_chain], the_table.name))
            if reference_chain:
                query += "    WHERE %s\n" % "\n          AND ".join([format_fk_join(fk) for fk in reference_chain])
                query += "          AND %s;" % pk
            else:
                query += "    WHERE %s;" % pk
            selects.append(query)

        selects = []
        create_select(the_table, info, selects, [], set())
        varsetup = []
        for c in the_table.primaryKey.columns:
            varsetup.append("SET @%s_to_select = <{row_id}>;" % c.referencedColumn.name)
        query = "\n".join(varsetup) + "\n" + "\n".join(selects)
        self.send(query)
        
        if not self.to_editor:
          mforms.App.get().set_status_text("Select statements for %s references was copied to clipboard" % the_table.name)


    def build_cascaded_delete(self):
        """
            To delete a row from a table all references to that row from other tables must be deleted first,
            otherwise the FK constraint won't let it be done.
            So this function will build the DELETE statements needed to do that, by deleting
            references to the given row all the way up to the wanted table.
        """
        editor = self.editor
        object = self.selection[0]

        info = dependencyInfoForSchemaInEditor(editor, object.schemaName)
        if not info:
            return
        info.scan_tables()
        the_table = info.table_with_name(object.schemaName, object.name)

        if not the_table:
            mforms.Utilities.show_error("Cascading DELETE", "Could not find reverse engineered %s" % object.name, "OK", "", "")
            return

        if not the_table.primaryKey:
            mforms.Utilities.show_error("Cascading DELETE", "Unable to create a DELETE statement for %s because the table has no primary key." % object.name,
                                        "OK", "", "")
            return

        # expression to match the row to be deleted
        pk = []
        for c in the_table.primaryKey.columns:
            pk.append("%s.%s = @%s_to_delete" % (the_table.name, c.referencedColumn.name, c.referencedColumn.name))
        pk = " AND ".join(pk)


        def create_delete(table, info, deletes, reference_chain, visited):
            if table in visited:
                return
            visited.add(table)

            def add_to_list(list, item):
                if item not in list:
                    return list + [item]
                return list

            for fk in info.get_foreign_keys_to_table(table):
                create_delete(fk.owner, info, deletes, reference_chain + [fk], visited)

            query = "DELETE FROM %s\n" % table.name
            query += "    USING %s\n" % ", ".join(add_to_list([fk.owner.name for fk in reference_chain], the_table.name))
            if reference_chain:
                query += "    WHERE %s\n" % "\n          AND ".join([format_fk_join(fk) for fk in reference_chain])
                query += "          AND %s;" % pk
            else:
                query += "    WHERE %s;" % pk
            deletes.append(query)

        deletes = []
        create_delete(the_table, info, deletes, [], set())
        varsetup = []
        for c in the_table.primaryKey.columns:
            varsetup.append("SET @%s_to_delete = <{row_id}>;" % c.referencedColumn.name)
        query = """
-- All objects that reference that row (directly or indirectly) will be deleted when this snippet is executed.
-- To preview the rows to be deleted, use Select Row Dependencies
START TRANSACTION;
-- Provide the values of the primary key of the row to delete.
%s

%s
COMMIT;
""" % ("\n".join(varsetup), "\n".join(deletes))

        self.send(query)

        if not self.to_editor:
          mforms.App.get().set_status_text("DELETE statement for %s was copied to clipboard" % the_table.name)
        



# Copyright (c) 2012, 2016, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import locale
import threading
from collections import deque

import grt
import mforms

import wizard_page_widget

os_icon_suffix = ""
import sys
if sys.platform=="darwin":
    os_icon_suffix = "_mac"


class WizardTaskError(Exception):
    pass


class WizardTask(mforms.Box):
    def __init__(self, owner, func, label):
        mforms.Box.__init__(self, True)
        self.set_managed()
        self.set_release_on_add()
        
        self.owner = owner
        self.label = label
        self.func = func
        
        self._icon = mforms.newImageBox()
        self._label = mforms.newLabel(label)
        
        self._enabled = True
        
        self.set_spacing(8)
        self.add(self._icon, False, True)
        self.add(self._label, True, True)

        self.reset()

    def set_text(self, text):
        self._label.set_text(text)

    def run(self):
        self.set_running()
        grt.log_info("Wizard", "Execute '%s'\n" % self.label)
        self.owner.send_info("%s..." % self.label)
        try:
            self.func()
        except grt.UserInterrupt:
            self.owner.send_info("%s cancelled" % self.label)
            self.set_aborted()
            return
        except Exception, e:
            if not isinstance(self, WizardThreadedTask):
                print
                import traceback
                traceback.print_exc()

            msg = "Error during %s: %s" % (self.label, self.owner.format_exception_text(e))
            self.owner.send_error(msg)
            #grt.log_error("Wizard", msg)
            self.set_failed(msg)
            raise e

        self.owner.send_info("%s done" % self.label)
        #grt.log_info("Wizard", "%s done\n" % self.label)
        self.set_finished()
            

    @property
    def is_enabled(self):
        return self._enabled

    @property
    def is_running(self):
        return self._running

    @property
    def is_finished(self):
        return self._finished

    @property
    def is_failed(self):
        return self._failed
    
    @property
    def is_aborted(self):
        return self._aborted
        
    @property
    def is_pending(self):
        return self.is_enabled and not self.is_finished and not self.is_failed and not self.is_aborted
    
    def reset(self):
        self._error_message = None
        self._running = False
        self._finished = False
        self._failed = False
        self._aborted = False
        self._label.set_text(self.label)
        if self._enabled:
            self._icon.set_image("task_unchecked%s.png" % os_icon_suffix)
        else:
            self._icon.set_image("task_disabled%s.png" % os_icon_suffix)
        
    def set_running(self):
        self._running = True
        self._icon.set_image("task_executing%s.png" % os_icon_suffix)

    def set_aborted(self):
        self._aborted = True
        self._running = False
        self._icon.set_image("task_error%s.png" % os_icon_suffix)
        
    def set_failed(self, msg):
        self._running = False
        self._failed = True
        self._error_message = msg
        self._icon.set_image("task_error%s.png" % os_icon_suffix)

    def set_error_icon(self):        
        self._icon.set_image("task_error%s.png" % os_icon_suffix)

    def set_warning_icon(self):        
        self._icon.set_image("task_warning%s.png" % os_icon_suffix)

    def set_finished(self):
        self._running = False
        self._finished = True
        self._icon.set_image("task_checked%s.png" % os_icon_suffix)

    def set_enabled(self, flag):
        if self._enabled != flag:
            self._enabled = flag
            if flag:
                self._icon.set_image("task_unchecked%s.png" % os_icon_suffix)
            else:
                self._icon.set_image("task_disabled%s.png" % os_icon_suffix)


class WizardThreadedTask(WizardTask):
    def __init__(self, owner, func, label):
        WizardTask.__init__(self, owner, func, label)
  
        self._thread = None
        
        
    class TaskThread(threading.Thread):
        def __init__(self, owner):
            threading.Thread.__init__(self)

            self.owner = owner
            
        def run(self):
            self.owner.thread_work()

        
  
    def run(self):
        self.set_running()
        self.owner.send_info("%s...." % self.label)
        grt.log_info("Wizard", "Starting thread for '%s'\n" % self.label)
        self._thread = self.TaskThread(self)
        self._thread.start()


    def reset(self):
        self._finished = False
        WizardTask.reset(self)


    def thread_work(self):
        try:
            self.func()
        except grt.UserInterrupt:
            self.owner.send_info("%s cancelled" % self.label)
            mforms.Utilities.perform_from_main_thread(self.set_aborted, False)
            return
        except Exception, exc:
            print
            import traceback
            traceback.print_exc()
            grt.log_error("Wizard", "Thread '%s' failed: %s\n" % (self.label, exc))
            self.owner.send_error("%s: %s" % (self.label, str(exc)))
            mforms.Utilities.perform_from_main_thread(lambda self=self,exc=exc:self.set_failed("Error: %s" % self.owner.format_exception_text(exc)), False)
            return

        self.owner.send_info("%s finished" % self.label)
        mforms.Utilities.perform_from_main_thread(self.set_finished, False)



class WizardProgressPage(wizard_page_widget.WizardPage):
    def __init__(self, main, header_label, description = None, use_private_message_handling=False):
        wizard_page_widget.WizardPage.__init__(self, main, header_label)
      
        self._use_private_message_handling = use_private_message_handling
      
        self._description = mforms.newLabel(description or "The following tasks will now be performed. Please monitor the execution.")
        self.content.add(self._description, False, True)
        
        self._tasks_box = mforms.newBox(False)
        self._tasks_box.set_padding(24)
        self._tasks_box.set_spacing(8)
        self.content.add(self._tasks_box, False, True)
        
        self._status_label = mforms.newLabel("Click [Next >] to execute.")
        self.content.add(self._status_label, False, True)
        
        self._progress = mforms.newProgressBar()
        self.content.add(self._progress, False, True)
        self._progress.show(False)
        
        self._detail_label = mforms.newLabel("")
        self.content.add(self._detail_label, False, True)

        self._timer = None
        self._tasks = []
        self._currently_running_task_index = None
        self._progress_indeterminate = False
        self._tasks_finished = False
        
        self._log_box = mforms.newPanel(mforms.TitledBoxPanel)
        self._log_box.set_title("Message Log")
        self._log_box.set_padding(12)
        
        self._log_text = mforms.newTextBox(mforms.VerticalScrollBar)
        self._log_text.set_name('WizardProgressLogText')
        self._log_text.set_read_only(True)
        self._log_box.add(self._log_text)
        self._log_box.show(False)
        self.content.add_end(self._log_box, True, True)
        self._log_queue = deque()
        
        self._showing_logs = False
        self._cancel_requested = False
        self._tasks_held = False
        self.advanced_button.set_text("Show Logs")
        
        self._log_progress_text = True
        self._autostart = False

    
    def clear_tasks(self):
        for task in self._tasks:
            self._tasks_box.remove(task)
        self._tasks = []
        self.reset()
        
    def format_exception_text(self, e):
        return str(e)    
        
    def add_task(self, func, label):
        """ Add a task to be executed from the main thread. The callback must be as in
          callback(progress), where progress is a function to be optionally executed to
          report progress. 
           
          The callback must return True if it finished executing or False if it was cancelled 
          by the user. If an error occurs during the execution of the task, an exception must 
          be raised.
        """
        wtask = WizardTask(self, func, label)
        self._tasks.append(wtask)
        self._tasks_box.add(wtask, False, True)
        return wtask
    
    
    def add_threaded_task(self, func, label):
        """ Add a task to be executed from a worker thread. The callback must be as in
          callback(). Progress information must be reported using grt.send_progress()
           
          The callback must return True if it finished executing or False if it was cancelled 
          by the user. If an error occurs during the execution of the task, an exception must 
          be raised.
        """
        wtask = WizardThreadedTask(self, func, label)
        self._tasks.append(wtask)
        self._tasks_box.add(wtask, False, True)
        return wtask


    def update_progress(self, pct, status):
        self._progress_info = (pct, status)


    def go_next(self):
        if not self._tasks_finished:
            if self._currently_running_task_index is None:
                self.start()
            else:
                self.resume()
        else:
            wizard_page_widget.WizardPage.go_next(self)

    def go_advanced(self):
        self._showing_logs = not self._showing_logs
        if self._showing_logs:
            self.advanced_button.set_text("Hide Logs")
        else:
            self.advanced_button.set_text("Show Logs")
        self._log_box.show(self._showing_logs)
    
    def go_cancel(self):
        self._cancel_requested = True

    def page_activated(self, advancing):
        super(WizardProgressPage, self).page_activated(advancing)
        
        self.next_button.set_enabled(self._tasks_finished or self._currently_running_task_index is None)

        if advancing and self._autostart:
            if self._currently_running_task_index is None and not self._tasks_finished:
                self.start()

    def reset(self, clear_log_box= False):
        self._cancel_requested = False
        self._tasks_finished = False
        self._progress_info = None
        self._currently_running_task_index = None
        self._progress.set_value(0)
        self._status_label.set_text("Click [Next >] to execute.")
        self._detail_label.set_text("")
        for task in self._tasks:
            task.reset()
        if clear_log_box:
            self._log_text.set_value("")

    def _handle_task_output(self, msg_type, text, detail):
        self._log_queue.append((msg_type, text, detail))
        if msg_type == "PROGRESS" and self._log_progress_text:
            text = text.split(":", 1)[-1]
            if detail:
                progress_text = "- %s: %s\n" % (text, detail)
            else:
                progress_text = "- %s\n" % (text)
            self._log_queue.append(("OUTPUT", progress_text, ""))

        return True

    def query_cancel_status(self):
        return self._cancel_requested

    def send_raw(self, msg):
        self._log_queue.append(("OUTPUT", msg, ""))

    def send_info(self, msg):
        grt.log_debug("Wizard", msg)
        self._handle_task_output("INFO", msg, "")
        
    def send_error(self, msg):
        grt.log_debug("Wizard", "ERROR: "+msg)
        self._handle_task_output("ERROR", msg, "")
        
    def send_warning(self, msg):
        grt.log_debug("Wizard", "WARNING: "+msg)
        self._handle_task_output("WARNING", msg, "")

    def send_progress(self, pct, msg):
        self._handle_task_output("PROGRESS", "%s:%s"%(locale.str(pct), msg), "")

    def start(self):
        self._warnings = 0
        self._errors = 0
        self._cancel_requested = False
        self._task_warnings = []
        self._task_errors = []
        self._progress_info = None
        self._currently_running_task_index = 0
        self._progress.show()
        self.next_button.set_enabled(False)
        self.back_button.set_enabled(False)
        self.cancel_button.set_enabled(True)
        
        if not self._use_private_message_handling:
            grt.push_status_query_handler(self.query_cancel_status)
            grt.push_message_handler(self._handle_task_output)
        self.send_info("Starting...")
        self._timer = mforms.Utilities.add_timeout(0.1, self.update_status)
        
    def resume(self):
        self._progress.show()
        self.next_button.set_enabled(False)
        self.back_button.set_enabled(False)
        self.cancel_button.set_enabled(True)
        
        if not self._use_private_message_handling:
            grt.push_status_query_handler(self.query_cancel_status)
            grt.push_message_handler(self._handle_task_output)
        self.send_info("Resuming...")

        self._tasks_held = False
        self._timer = mforms.Utilities.add_timeout(0.1, self.update_status)

        
    def tasks_finished(self):
        pass


    def tasks_failed(self, canceled):
        pass
        
    def _hold(self):
        self.send_info("Processing Held")
        self._flush_messages()
        self.send_info("\n")

        self._progress.show(False)
        self._progress.stop()
        self._progress_indeterminate = False
        self.next_button.set_enabled(True)
        self.back_button.set_enabled(True)
        self.cancel_button.set_enabled(True)

        self._tasks_held = True



    def _finished(self):
        if self._warnings > 0 or self._errors > 0:
            self.send_info("Tasks finished with warnings and/or errors; view the logs for details")
        self.send_info(self.final_message())
        self._flush_messages()
        self.send_info("\n\n")
        if not self._use_private_message_handling:
            grt.pop_message_handler()
            grt.pop_status_query_handler()
    
        self._progress.show(False)
        self._progress.stop()
        self._progress_indeterminate = False
        self._tasks_finished = True
        self.next_button.set_enabled(True)
        self.back_button.set_enabled(True)
        self.cancel_button.set_enabled(False)
        
        self._status_label.set_text("")
        
        # Updates the details label as needed
        self._detail_label.set_text(self.final_details())

        if self._errors > 0 or self._warnings > 0:
            if not self._showing_logs:
                self.go_advanced()
            self.tasks_failed(False)
        else:
            self.tasks_finished()


    def _failed(self):
        self.send_info("Failed")
        self._flush_messages()
        self.send_info("\n")
        if not self._use_private_message_handling:
            grt.pop_message_handler()
            grt.pop_status_query_handler()

        self._progress.show(False)
        self._progress.stop()
        self._progress_indeterminate = False
        self.next_button.set_enabled(False)
        self.back_button.set_enabled(True)
        self.cancel_button.set_enabled(False)
        
        self._detail_label.set_text(self.failure_message())

        if not self._showing_logs:
            self.go_advanced()

        self.tasks_failed(False)
        

    def _cancelled(self):
        self.send_info("Cancelled")
        self._flush_messages()
        self.send_info("\n")
        if not self._use_private_message_handling:
            grt.pop_message_handler()
            grt.pop_status_query_handler()

        self._progress.show(False)
        self._progress.stop()
        self._progress_indeterminate = False
        self.next_button.set_enabled(False)
        self.back_button.set_enabled(True)
        self.cancel_button.set_enabled(False)
        
        self._detail_label.set_text("Cancelled by user.")

        self.tasks_failed(True)


    def _flush_messages(self):
        while len(self._log_queue) > 0:
            mtype, text, detail  = self._log_queue.popleft()
            if "\n" in detail:
                detail = "\n    "+("\n    ".join(detail.split("\n")))
            if mtype == "INFO":
                if detail:
                    self._log_text.append_text_and_scroll("%s: %s\n" % (text, detail), True)
                else:
                    self._log_text.append_text_and_scroll("%s\n" % (text), True)
            elif mtype == "OUTPUT":
                # output text must already come with newlines
                if detail:
                    self._log_text.append_text_and_scroll("%s: %s" % (text, detail), True)
                else:
                    self._log_text.append_text_and_scroll("%s" % (text), True)
            elif mtype == "PROGRESS":
                pct, y, text = text.partition(":")
                try:
                    self.update_progress(locale.atof(pct), text)
                except Exception:
                    grt.log_debug('Wizard', 'Exception raised when converting "%s" to float using locale.atof(). Exception ignored\n' % pct)
            else:
                if mtype == "ERROR":
                    self._errors += 1
                    self._task_errors.append(text)
                elif mtype == "WARNING":
                    self._warnings += 1
                    self._task_warnings.append(text)
                if detail:
                    self._log_text.append_text_and_scroll("%s: %s: %s\n" % (mtype, text, detail), True)
                else:
                    self._log_text.append_text_and_scroll("%s: %s\n" % (mtype, text), True)

    def failure_message(self):
        """Subclass and override to change the text message to be shown when tasks failed to finish."""
        return "A task has failed executing."


    def final_message(self):
        """Subclass and override to change the text message to be shown when tasks finish successfully."""
        return "Finished performing tasks."


    def final_details(self):
        """Subclass and override to change the text message to be shown when tasks finish successfully."""
        ret_val = ""
        if self._errors > 0:
            ret_val = "There were errors during execution; please review log messages."
        elif self._warnings > 0:
            ret_val = "There were warnings during execution; please review log messages.\nClick [Next >] to continue if you think they are not important."
        else:
            ret_val = self.final_message() + "\nClick [Next >] to continue."

        return ret_val

    def verify_task_preconditions(self, label):
        """
        Will check for specific task conditions to be met, task will be identified by its label.
        If preconditions are met returns None, if not, returns a descriptive error message.
        """
        return None

    def update_status(self):
        self._flush_messages()

        if self._tasks_held:
            return False
    
        if self._progress_info:
            pct, text = self._progress_info
            self._progress_info = None
            if pct < 0:
                if not self._progress_indeterminate:
                    self._progress_indeterminate = True
                    self._progress.set_indeterminate(True)
                    self._progress.start()
            else:
                if self._progress_indeterminate:
                    self._progress_indeterminate = False
                    self._progress.stop()
                    self._progress.set_indeterminate(False)
                self._progress.set_value(pct)
            self._status_label.set_text(text)
    
        task = self._tasks[self._currently_running_task_index]

        if not task.is_running and task.is_pending:
            if self._cancel_requested:
                self._cancelled()
                return False

            try:
                error = self.verify_task_preconditions(task.label)
                if error is None:
                    task.run()
                else:
                    self._hold()
                    mforms.Utilities.show_error("Unable To Continue", error, "OK", "", "")


            except Exception, exc:
                #print
                #import traceback
                #traceback.print_exc()
                self.send_error("Exception in task '%s': %r\n" % (task.label, exc))
                self._status_label.set_text("%s" % self.format_exception_text(exc))
                self._failed()
                return False
    
        if task.is_failed:
            self._failed()
            return False
    
        if task.is_aborted:
            self._cancelled()
            return False
        
        if task.is_finished or not task.is_enabled: # go to next task
            if self._task_errors:
                task.set_error_icon()
            elif self._task_warnings:
                task.set_warning_icon()
            self._task_errors = []
            self._task_warnings = []
            self._currently_running_task_index += 1

            if self._currently_running_task_index == len(self._tasks):
                # last task was executed
                self._finished()
                return False

        return True



# Copyright (c) 2012, 2014, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import grt
from grt.modules import Workbench
import mforms
from wb import wbinputs
from wb_utils_grt import ModuleInfo

@ModuleInfo.plugin('wb.util.copySQLToClipboard', caption='Copy SQL to Clipboard', input= [wbinputs.objectOfClass('db.DatabaseObject')], groups= ['Catalog/Utilities', 'Menu/Objects'])
@ModuleInfo.export(grt.INT, grt.classes.GrtNamedObject)
def copySQLToClipboard(obj):
        script = []        
        # workaround until diff sql generator handles routine groups
        if isinstance(obj, grt.classes.db_RoutineGroup):
            for routine in obj.routines:
                script.append(grt.modules.DbMySQL.makeCreateScriptForObject(routine))
                script.append(';\n\n')
        else:
            script.append(grt.modules.DbMySQL.makeCreateScriptForObject(obj))

        Workbench.copyToClipboard(''.join(script))
        return 0
          
@ModuleInfo.plugin('wb.util.copyColumnNamesToClipboard', caption='Copy Column Names to Clipboard', input= [wbinputs.objectOfClass('db.Table')], groups= ['Catalog/Utilities', 'Menu/Objects'])
@ModuleInfo.export(grt.INT, grt.classes.db_Table)
def copyColumnNamesToClipboard(table):
        data = ', '.join([column.name for column in table.columns])
        Workbench.copyToClipboard(data)
        return 0

@ModuleInfo.plugin('wb.util.copyTableListToClipboard', caption='Copy Table List to Clipboard', input= [wbinputs.currentCatalog()], groups= ['Catalog/Utilities', 'Menu/Catalog'])
@ModuleInfo.export(grt.INT, grt.classes.db_Catalog)
def copyTableListToClipboard(cat):    
    #insert = ['`'+schema.name+'`.`'+tbl.name+'`' for tbl in schema.tables for schema in cat.schemata ]
    insert = '' 
    for schema in cat.schemata:
        insert = insert + ', '.join(['`'+schema.name+'`.`'+tbl.name+'`' for tbl in schema.tables])
             
    Workbench.copyToClipboard(insert)
    return 0

def generateName(name_prefix, names_map):
     name_suffix = 0
     byte_a = ord('a')
     while True:
         for i in range(0, 25):
             next_name = ''
             if name_prefix == '':
                 next_name = chr(byte_a + i) + str(name_suffix)
             else:
                 next_name = name_prefix + str(name_suffix)
             
             if not next_name in names_map:
                 names_map[next_name] = 1
                 return next_name

         name_suffix = name_suffix + 1
         
@ModuleInfo.plugin('wb.util.obfuscateCatalog', caption='Obfuscate Object Names in Catalog', input= [wbinputs.currentModel()], groups= ['Catalog/Utilities', 'Menu/Utilities'])
@ModuleInfo.export(grt.INT, grt.classes.workbench_physical_Model)
def obfuscateCatalog(model):     
 
     result = mforms.Utilities.show_warning('Warning', '''This operation will change names of all schemata and tables in the model.
Would you like to also clear SQL code for views and stored procedures?
This action cannot be undone!''', 'Clear SQL', 'Cancel', 'Leave SQL')

     if result == mforms.ResultCancel:
         return 1

     clear_sql_code = (result == mforms.ResultOk)
     schemata_map = {}
     objects_map = {}
     
     def createFiguresMap(model):
         figures_map = {}
         for diagram in model.diagrams:        
             for figure in diagram.figures:
                 if hasattr(figure, 'table'):
                     figures_map[figure.table.__id__] = figure
                 elif hasattr(figure, 'view'):
                     figures_map[figure.view.__id__] = figure
                 elif hasattr(figure, 'routine'):
                     figures_map[figure.routine.__id__] = figure
                 elif hasattr(figure, 'routineGroup'):
                     figures_map[figure.routineGroup.__id__] = figure
         return figures_map
    
     figures_map = createFiguresMap(model)
     
     def renameFigure(obj, figures_map):
         if obj.__id__ in figures_map:
             figures_map[obj.__id__].name = obj.name
     
     def renameDbObject(obj, obj_map=objects_map, clear_sql_code=clear_sql_code, figures_map=figures_map):
         obj.name = generateName('', obj_map)
         if clear_sql_code and hasattr(obj, 'sqlBody'):
             obj.sqlBody = ''
         if clear_sql_code and hasattr(obj, 'sqlDefinition'):
             obj.sqlDefinition = ''
         renameFigure(obj, figures_map)                      
        
     for schema in model.catalog.schemata:
        schema.name = generateName('', schemata_map)        
        for table in schema.tables:
            renameDbObject(table)
        for view in schema.views:
            renameDbObject(view)
        for routine in schema.routines:
            renameDbObject(routine)
        for routineGroup in schema.routineGroups:
            renameDbObject(routineGroup)
            for routine in routineGroup.routines:
                renameDbObject(routine)
   
     return 0

@ModuleInfo.plugin('wb.util.prefixTables', caption='Give a Prefix to All Tables in Catalog', input= [wbinputs.currentCatalog()], groups= ['Catalog/Utilities', 'Menu/Catalog'])
@ModuleInfo.export(grt.INT, grt.classes.db_Catalog)
def prefixTables(cat):
     
     ret, prefix = mforms.Utilities.request_input("Give a Prefix to All Tables in Catalog", "Please specify the prefix:copy table ", "")
     
     if not ret:
         return 1

     for schema in cat.schemata:
         for tbl in schema.tables:
             tbl.name = prefix + tbl.name
             
     return 0             

@ModuleInfo.plugin('wb.util.changeStorageEngines', caption='Change the Storage Engine of All Tables', input= [wbinputs.currentCatalog()], groups= ['Catalog/Utilities', 'Menu/Catalog'])
@ModuleInfo.export(grt.INT, grt.classes.db_Catalog)
def changeStorageEngines(cat):
     ret, new_engine = mforms.Utilities.request_input("Change the Storage Engine of All Tables", "Type the new storage engine name:", "")
     
     if not ret:
         return 1
 
     def getTableEngines():
        result = grt.root.wb.options.options['@db.mysql.Table:tableEngine/Items']
        items = [item.strip(' \t') for item in result.split(',')]
        return items

     # validate the engine name and fix its case
     engines = getTableEngines()
     engine_found = False
     for engine_name in engines:
        if engine_name.find(':') != -1:
            engine_name = engine_name[engine_name.find(':') + 1:]
        if new_engine.lower() == engine_name.lower():
            engine_found = True
            new_engine = engine_name

     if not engine_found:
        Workbench.confirm('Change Storage Engines', 'Invalid storage engine name: ' + new_engine)
        return 2

     for schema in cat.schemata:
         for tbl in schema.tables:
            tbl.tableEngine = new_engine
            
     return 0



# Copyright (c) 2012, 2016, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import re
from grt import log_error

def get_exe_path(cmd):
    import os
    import mforms

    filepath = mforms.App.get().get_executable_path(cmd).encode("utf8")
    if len(filepath) != 0:
        return filepath;
    
    filepath, filename = os.path.split(cmd)
    
    def is_executable(filepath):
        return os.path.isfile(filepath) and os.access(filepath, os.X_OK)
    
    if filepath:
        if is_executable(filepath):
            return cmd
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe = os.path.join(path, cmd)
            if is_executable(exe):
                return exe
    return None

def human_size(num):
    for x in ['bytes','KiB','MiB','GiB']:
        if num < 1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0
    return "%3.1f %s" % (num, 'TiB')


def format_duration(t, skip_seconds=False):
    s = int(t % 60)
    m = int((t / 60) % 60)
    h = int((t / 3600) % 24)
    d = int(t / (3600*24))

    res = []
    if d > 0:
        if d == 1:
            res.append("%i day" % d)
        else:
            res.append("%i days" % d)

    if skip_seconds:
        res.append("%i:%02i" % (h, m))
    else:
        res.append("%i:%02i:%02i" % (h, m, s))

    return " ".join(res)



def find_object_with_name(list, name):
    """Finds an object with the given name within a list of objects (such as grt.List).

    Returns the found object or None if there was no object with the given name in the collection.
    """
    for obj in list:
        if obj.name == name:
            return obj
    return None

def find_object_with_old_name(list, name):
    """Finds an object with the given oldName within a list of objects (such as grt.List).

    Returns the found object or None if there was no object with the given name in the collection.
    """
    for obj in list:
        if obj.oldName == name:
            return obj
    return None


def replace_string_parameters(template_string, params):
    if isinstance(params, dict):
        params = list( params.iteritems() )
    return reduce( lambda partial_template_string, rep_tuple: partial_template_string.replace('%'+rep_tuple[0]+'%', str(rep_tuple[1])),
                                [ template_string ] + params
                  )

def parameters_from_dsn(dsn):
    chunks = dsn.split(';')
    params = ( (name, value) for name, value in 
                    ( chunk.split('=', 1) for chunk in chunks if '=' in chunk )
                    if not (value.startswith('%') and value.endswith('%'))
             )
    return dict(params)

def dsn_parameters_to_connection_parameters(dsn_params):
    param_mapping = { 'DRIVER'  : 'driver',
                      'SERVER'  : 'hostName',
                      'UID'     : 'userName',
                      'PWD'     : 'password',
                      'PORT'    : 'port',
                      'DATABASE': 'schema',
                      'DSN'     : 'dsn',
                    }
    return dict( (param_mapping.get(dsn_key.upper(), dsn_key), dsn_value) for dsn_key, dsn_value in dsn_params.iteritems() )
    
    
def check_grt_subtree_consistency(value):
    pass

def server_version_str2tuple(version_str):
    match = re.match(r'^(\d+\.\d+(\.\d+)*).*$', version_str.strip())
    if match:
        return tuple(int(x) for x in match.group(1).split('.'))
    return tuple()

def server_os_path(server_profile):
    """Returns an os.path module specific for the server OS."""
    if server_profile.target_is_windows:
        return __import__('ntpath')
    else:
        return __import__('posixpath')



class Version:
    def __init__(self, major, minor, release=0):
        self.majorNumber = major
        self.minorNumber = minor
        self.releaseNumber = release

    def __str__(self):
        if self.releaseNumber >= 0:
            return "%i.%i.%i" % (self.majorNumber, self.minorNumber, self.releaseNumber)
        else:
            return "%i.%i" % (self.majorNumber, self.minorNumber)

    @classmethod
    def fromgrt(cls, v):
        return Version(v.majorNumber, v.minorNumber, v.releaseNumber)

    @classmethod
    def fromstr(cls, s):
        match = re.match(r'^(\d+\.\d+(\.\d+)*).*$', s.strip())
        if match:
            v = tuple(int(x) for x in match.group(1).split('.'))
        else:
            v = []

        if len(v) == 1:
            return Version(v[0])
        elif len(v) == 2:
            return Version(v[0], v[1])
        elif len(v) == 3:
            return Version(v[0], v[1], v[2])
        else:
            raise ValueError("Invalid version string %s" % s)
    
    def compare(self, other):
        other_version = None
        if isinstance(other, Version):
            other_version = other
        elif isinstance(other, basestring):
            other_version = Version.fromstr(other)
        else:
            raise TypeError("Unexpected type")

        this_version_number = self.majorNumber * 10000 + self.minorNumber * 100 + max(0, self.releaseNumber)
        other_version_number = other_version.majorNumber * 10000 + other_version.minorNumber * 100 + max(0, other_version.releaseNumber)
        
        if this_version_number < other_version_number:
            return -1
        elif this_version_number > other_version_number:
            return 1
        return 0
    
    def __lt__(self, other):
        return self.compare(other) < 0

    def __eq__(self, other):
        return self.compare(other) == 0

    def __ne__(self, other):
        return self.compare(other) != 0

    def __gt__(self, other):
        return self.compare(other) > 0

    def __ge__(self, other):
        return self.compare(other) >= 0

    def __le__(self, other):
        return self.compare(other) <= 0

    def is_supported_mysql_version(self):
        if (self.majorNumber == 5 and self.minorNumber in (1, 5, 6, 7)) or (self.majorNumber == 8 and self.minorNumber == 0):
            return True
        return False

    def is_supported_mysql_version_at_least(self, major, minor = None, release=-1):
        assert type(major) == int or isinstance(major, Version)
        if isinstance(major, Version):
            v = major
            major = v.majorNumber
            minor = v.minorNumber
            release = v.releaseNumber

        # if the version required is older (<) than 5.6, then any server that matches is fine
        # if the version required is newer (>=) than 5.6, then we can only guarantee that known servers versions have some specific feature
        
        if (major == 5 and minor >= 6) or (major == 8 and minor == 0):
            return self.is_supported_mysql_version() and self >= Version(major, minor, release)
        else:
            return self > Version(major, minor, release)

import threading
import Queue

class QueueFile:
    def __init__(self):
        self._cond = threading.Condition()
        self.data = ""
        self._write_done = False

    def write(self, data):
        self._cond.acquire()
        self.data += data
        self._cond.notify()
        self._cond.release()

    def close(self):
        self._cond.acquire()
        self._write_done = True
        self._cond.notify()
        self._cond.release()

    def peek(self, size):
        data = ""
        self._cond.acquire()
        while size > len(self.data) and not self._write_done:
            self._cond.wait()
        if self._write_done:
            data = self.data
        else:
            data = self.data[:size]
        self._cond.release()
        return data

    def read(self, size):
        data = ""
        self._cond.acquire()
        while size > len(self.data) and not self._write_done:
            self._cond.wait()
        if self._write_done:
            data = self.data
            self.data = ""
        else:
            data = self.data[:size]
            self.data = self.data[size:]
        self._cond.release()
        return data

    def readline(self):
        data = ""
        self._cond.acquire()
        find_start = 0
        # in case of VERY long lines (several MBs) this loop can slow down things a lot, so we take a longer break every once in a while to
        # give the thread more time to feed data
        while self.data.find('\n', find_start) < 0 and not self._write_done:
            find_start = len(self.data)
            self._cond.wait()
        pos = self.data.find('\n')
        if self._write_done:
            if pos >= 0:
                pos += 1
                data = self.data[:pos]
                self.data = self.data[pos:]
            else:
                data = self.data
                self.data = ""
        else:
            pos += 1
            data = self.data[:pos]
            self.data = self.data[pos:]
        self._cond.release()
        return data


import multiprocessing
class QueueFileMP:
    def __init__(self, pipe):
        self._queue = pipe
        self._write_done = False
        self._data = ""

    def write(self, data):
        self._queue.put(data)

    def close(self):
        self._queue.put(None)

    def _readup(self, maxloops=4):
        tmp = self._queue.get()
        if tmp is None:
            self._write_done = True
        else:
            l = [tmp]
            # flush the queue
            for i in range(maxloops):
                try:
                    tmp = self._queue.get()
                    if tmp is None:
                        self._write_done = True
                        break
                    l.append(tmp)
                except multiprocessing.Queue.Empty:
                    break
            self._data += "".join(l)

    def peek(self, size):
        while size > len(self._data) and not self._write_done:
            self._readup(0)
        if self._write_done:
            data = self._data
        else:
            data = self._data[:size]
        return data

    def read(self, size):
        data = ""
        while size > len(self._data) and not self._write_done:
            self._readup()
        if self._write_done:
            data = self._data
            self._data = ""
        else:
            data = self._data[:size]
            self._data = self._data[size:]
        return data

    def readline(self):
        data = ""
        find_start = 0
        # in case of VERY long lines (several MBs) this loop can slow down things a lot, so we take a longer break every once in a while to
        # give the thread more time to feed data
        while self._data.find('\n', find_start) < 0 and not self._write_done:
            find_start = len(self._data)
            self._readup()
        pos = self._data.find('\n')
        if self._write_done:
            if pos >= 0:
                pos += 1
                data = self._data[:pos]
                self._data = self._data[pos:]
            else:
                data = self._data
                self._data = ""
        else:
            pos += 1
            data = self._data[:pos]
            self._data = self._data[pos:]
        return data



class WorkerThreadHelper:
    """Worker thread that executes a task and sends messages/updates to a message handler 
        that keeps running from the main thread until it's done."""
    def __init__(self, worker_func, message_handler_func):
        self.worker = worker_func
        self.message_handler = message_handler_func
        self.thread = threading.Thread()
        self.queue = Queue.Queue()
        self.thread.run = self._run
        self._timeout_handle = None
        self._running = False


    def start(self, update_interval = 0.5):
        self.thread.start()
        self._running = True
        import mforms
        self._timeout_handle = mforms.Utilities.add_timeout(update_interval, self._timeout)

    def add_message(self, message):
        self.queue.put(message)

    def _timeout(self):
        while True:
            try:
                message = self.queue.get_nowait()
            except Queue.Empty:
                break
            self.message_handler(message)
        return self._running and self.queue.empty()

    def _run(self):
        try:
            self.worker(self.add_message)
        except Exception, e:
            import traceback
            log_error("WorkerThreadHelper", "An exception occurred in the worker thread:\n%s\n" % traceback.format_exc())
            self.add_message(e)
        self._running = False




# Copyright (c) 2014, 2015 Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA


import mforms

class WizardForm(mforms.Form):
    def __init__(self, owner_form):
        mforms.Form.__init__(self, owner_form)

        self.pages = []

        self.content = mforms.newBox(False)
        self.content.set_spacing(12)

        hbox = mforms.newBox(True)
        self.header = mforms.newLabel("")
        self.header.set_style(mforms.WizardHeadingStyle)
        hbox.add(self.header, True, True)
        hbox.set_padding(24)

        self.content.add(hbox, False, True)
        self.tabview = mforms.newTabView(mforms.TabViewTabless)
        self.content.add(self.tabview, True, True)

        self.set_content(self.content)
        self.set_size(800, 600)
        self.center()
        
        self.set_on_close(self.on_close)
     
    def on_close(self):
        current_page = self.tabview.get_active_tab()
        if hasattr(self.pages[current_page], "on_close"):
            return self.pages[current_page].on_close()
        return True


    def add_page(self, page):
        if self.pages:
            self.pages[len(self.pages) - 1].set_last_page(False)
        else:
            page.back_button.set_enabled(False)
        page.set_last_page(True)
        self.pages.append(page)
        self.tabview.add_page(page, "")


    def go_next_page(self):
        current_page = self.tabview.get_active_tab()
        if not self.pages[current_page].validate():
            return
        if current_page == len(self.pages) - 1:
            self.finish()
        for index in range(current_page + 1, len(self.pages)):
            if self.pages[index].should_skip():
                continue
            self.tabview.set_active_tab(index)
            self.pages[index].page_activated(True)
            break


    def go_previous_page(self):
        current_page = self.tabview.get_active_tab()
        for index in range(current_page - 1, -1, -1):
            if self.pages[index].should_skip():
                continue
            self.tabview.set_active_tab(index)
            self.pages[index].page_activated(False)
            break

    def finish(self):
        self.close()

    def cancel(self):
        self.end_modal(False)


    def run(self, modal = False):
        self.pages[0].page_activated(True)

        if modal:
            self.run_modal(None, None)
        else:
            self.show()






# Copyright (c) 2012, 2016, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import mforms
from workbench.ui import WizardPage

class MainView(WizardPage):
    def __init__(self, main):
        WizardPage.__init__(self, main, 'Overview', no_buttons=True)

        self.main = main

    def create_ui(self):
        label = mforms.newLabel('Welcome to the MySQL Workbench Schema Transfer Wizard')
        label.set_style(mforms.BigBoldStyle)
        self.content.add(label, False, True)
        self.content.set_spacing(12)
        self.content.set_padding(20)

        label = mforms.newLabel('The MySQL Schema Transfer Wizard helps you to move your data from an older MySQL server to the latest MySQL GA\n'+
                                '(General Availability) version. It is meant for developer machines to get you working with the latest\n'+ 
                                'MySQL Server quickly. The data is transferred on the fly and is not based on a consistent snapshot. This\n'+
                                'works well for local instances that are used for development purposes. Please note that you should not\n'+
                                'use this tool on production MySQL instances. Production databases require a more complex data migration\n'+
                                'scenario in most cases.')
        self.content.add(label, False, True)

        box = mforms.newBox(True)
        box.add(mforms.newLabel(''), True, True)
        button_start = mforms.newButton()
        button_start.set_size(150, -1)
        button_start.set_text('Start the Wizard')
        button_start.add_clicked_callback(self.main.go_next_page)
        box.add(button_start, False, False)
        box.add(mforms.newLabel(''), True, True)
        self.content.add_end(box, False, True)



# Copyright (c) 2007, 2012, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA
import mforms
from wb_server_management import wbaOS

#===============================================================================
#
#===============================================================================
class UIProfile(object):
    def __init__(self, server_profile):
        self.server_profile = server_profile
        self.host_os = server_profile.host_os
        self.styles = {
                        wbaOS.windows :
                             {
                               'main' : (lambda x: x.set_back_color("#293852"),
                                         lambda x: x.set_spacing(6)
                                        ),
                               'page' : (lambda x: x.set_back_color("#ffffff"),),
                               'sidebar-label' : (lambda x: x.set_back_color("#4a6184"),
                                                  lambda x: x.set_color("#ffffff"),
                                                  lambda x: x.set_size(-1, 25)
                                                 ),
                               'content-label' : (lambda x: x.set_back_color("#4a6184"),
                                                  lambda x: x.set_color("#ffffff"),
                                                  lambda x: x.set_size(-1, 25)
                                                 ),
                               'subsection-label' : (lambda x: x.set_back_color("#ffffff"),
                                                     lambda x: x.set_color("#000000")
                                                    ),
                               'option-search-panel' : (lambda x: x.set_back_color("#bdc7de"),
                                                        lambda x: x.set_size(-1, 26)
                                                       )
                             },
                        wbaOS.linux :
                             {
                               'main' : (lambda x: x.set_padding(0),
                                         lambda x: x.set_spacing(1)
                                        ),
                               'page' : (lambda x: None,),
                               'sidebar-label' : (lambda x: x.set_color("#ffffff"),
                                                  lambda x: x.set_style(mforms.BoldStyle),
                                                  lambda x: x.set_size(-1, 24),
                                                 ),
                               'content-label' : (lambda x: x.set_color("#ffffff"),
                                                  lambda x: x.set_style(mforms.BoldStyle),
                                                  lambda x: x.set_size(-1, 20)
                                                 ),
                               'subsection-label' : (lambda x: x.set_back_color("#d9e2ef"),
                                                     lambda x: x.set_color("#000000")
                                                    ),
                               'option-search-panel' : (
                                                        lambda x: x.set_size(-1, 26),
                                                       )
                             },
                        wbaOS.darwin :
                             {
                               'main' : (lambda x: x.set_spacing(1),
                                         ),
                               'page' : (lambda x: None, ),
                               'sidebar-label' : (lambda x: x.set_back_color("#efefef"),
                                                  lambda x: x.set_color("#454545"),
                                                  lambda x: x.set_size(-1, 25)
                                                 ),
                               'content-label' : (lambda x: x.set_color("#454545"),
                                                  lambda x: x.set_style(mforms.SmallBoldStyle),
                                                  lambda x: x.set_text_align(mforms.MiddleCenter),
                                                  lambda x: x.set_size(-1, 22)
                                                 ),
                               'subsection-label' : (lambda x: x.set_style(mforms.SmallBoldStyle),
                                                     lambda x: x.set_color("#484950")
                                                    ),
                               'option-search-panel' : (lambda x: x.set_back_color("#f1f1f1"),
                                                        lambda x: x.set_size(-1, 26)
                                                       )
                             }
                      }

    #---------------------------------------------------------------------------
    def apply_style(self, target, style_name):
        style = None
        os_ui_profile = self.styles.get(self.host_os, None)
        if os_ui_profile:
            style = os_ui_profile.get(style_name)

        if style is not None:
            for style_part in style:
                style_part(target)
        else:
            print "OS profile has no style '%s'" % style_name



# Copyright (c) 2013, 2016, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA


import mforms
import grt

import re
import os

from threading import Thread
from Queue import Queue, Empty
from workbench.client_utils import MySQLScriptImporter

from wb_admin_utils import make_panel_header, MessageButtonPanel

from workbench.log import log_info, log_error




#===============================================================================
# WB SYS Schema Deployment

def get_current_sys_version(server_version):
    """Gets the version of the sys schema that's shipped with Workbench."""
    syspath = mforms.App.get().get_resource_path("sys")
    path = os.path.join(syspath, "before_setup.sql")
    if not os.path.exists(path):
        log_info("No sys script found\n")
        return None
    for line in open(path):
        if line.startswith("CREATE OR REPLACE ALGORITHM"):
            m = re.findall("SELECT '(.*)' AS sys_version", line)
            if m:
                return m[0]
    return None


def get_installed_sys_version(sql_editor):
    """Checks whether the sys schema is installed and up to date."""
    try:
        res = sql_editor.executeManagementQuery("SELECT sys_version FROM sys.version", 0)
        if res.goToFirstRow():
            return res.stringFieldValue(0)
    except grt.DBError, e:
        log_error("MySQL error getting sys schema version: %s\n" % e)
        if e.args[1] == 1146: # table doesn't exist
            return None
        raise



class HelperInstallPanel(mforms.Table):
    def __init__(self, owner, editor):
        mforms.Table.__init__(self)
        self.set_managed()
        self.set_release_on_add()

        self.set_row_count(2)
        self.set_column_count(1)

        self.set_padding(-1)

        self.label = mforms.newLabel("Installing...")
        self.add(self.label, 0, 1, 0, 1, mforms.HFillFlag)
        
        self.progress = mforms.newProgressBar()
        self.add(self.progress, 0, 1, 1, 2, mforms.HFillFlag)
        
        self.progress.set_size(400, -1)
        
        self.owner = owner
        
        self.importer = MySQLScriptImporter(editor.connection)
        self.importer.report_progress = self.report_progress
        self.importer.report_output = self.report_output
        
        self._worker_queue = Queue()
        self._worker = None
        self._progress_status = None
        self._progress_value = 0

        self._update_timer = None
        self._messages = []

    def __del__(self):
        if self._update_timer:
            mforms.Utilities.cancel_timeout(self._update_timer)

    @property
    def is_busy(self):
        return self._worker != None


    def report_progress(self, status, current, total):
        self._progress_status = status
        if total > 0:
            self._progress_value = float(current) / total


    def report_output(self, message):
        log_info("%s\n" % message)
        self._messages.append(message)


    def update_ui(self):
        try:
            data = self._worker_queue.get_nowait()
            if data is None:
                self._worker.join()
                self._worker = None
                self.owner.page_activated()
                self._update_timer = None
                if self._messages:
                    mforms.Utilities.show_message("Install sys Schema", "Import output:\n%s" % "\n".join(self._messages), "OK", "", "")
                return False

            if isinstance(data, Exception):
                if isinstance(data, grt.DBError) and data.args[1] == 1044:
                    mforms.Utilities.show_error("Install sys Schema",
                                                "The current MySQL account does not have enough privileges to create the sys schema.\nPlease use an account with schema creation privileges or ask an administrator to install sys.", "OK", "", "")
                elif isinstance(data, grt.DBLoginError):
                    mforms.Utilities.show_error("Install sys Schema",
                                                "Error installing sys Schema.\n"+str(data), "OK", "", "")
                else:
                    mforms.Utilities.show_error("Install sys Schema",
                                                "Error installing sys Schema.\n"+str(data), "OK", "", "")
        except Empty:
            pass

        if self._progress_status is not None:
            self.label.set_text(self._progress_status)
        self.progress.set_value(self._progress_value)

        return True
            
            
    def work(self, files):
        try:
            for f, db in files:
                log_info("Installing %s...\n" % f)
                self._progress_status = "Installing %s..." % f
                self._progress_value = 0
                self.importer.import_script(f, db)

            log_info("sys schema installation finished\n")
        except grt.DBLoginError, e:
            log_error("MySQL login error installing sys schema: %s\n" % e)
            self._worker_queue.put(e)
        except grt.DBError, e:
            log_error("MySQL error installing sys schema: %s\n" % e)
            self._worker_queue.put(e)
        except Exception, e:
            import traceback
            log_error("Unexpected exception installing sys schema: %s\n%s\n" % (e, traceback.format_exc()))
            self._worker_queue.put(e)
        self._worker_queue.put(None)


    def start(self):
        server_profile = self.owner.ctrl_be.server_profile
        parameterValues = server_profile.db_connection_params.parameterValues
        pwd = parameterValues["password"]
        if not pwd:
            username = parameterValues["userName"]
            host = server_profile.db_connection_params.hostIdentifier
            accepted, pwd = mforms.Utilities.find_or_ask_for_password("Install sys Schema", host, username, False)
            if not accepted:
                return
        self.importer.set_password(pwd)
      
        syspath = mforms.App.get().get_resource_path("sys")

        server_version = self.owner.ctrl_be.target_version
        main_sys = "sys_%i%i.sql" % (server_version.majorNumber, server_version.minorNumber)

        self._worker = Thread(target = self.work, args = ([(os.path.join(syspath, main_sys), None)],))
        self._worker.start()
    
        self._update_timer = mforms.Utilities.add_timeout(0.2, self.update_ui)



    def install(self, what, default_db):
        self.importer.import_script(what, default_db)



class WbAdminPSBaseTab(mforms.Box):
    sys = "sys"
    ui_created = False

    def __init__(self, ctrl_be, instance_info, main_view):
        mforms.Box.__init__(self, False)
        self.set_managed()
        self.set_release_on_add()
        
        self.set_spacing(12)
        self.set_padding(12)

        self.instance_info = instance_info
        self.ctrl_be = ctrl_be
        self.main_view = main_view
    
        self.installer_panel = None
        self.warning_panel = None
        self.content = None
        


    def create_basic_ui(self, icon, title, button=None):
        self.heading = make_panel_header(icon, self.instance_info.name, title, button)
        self.heading.set_padding(8)
        self.add(self.heading, False, False)

        self.content = None


    def get_select_int_result(self, query):
        res = self.main_view.editor.executeManagementQuery(query, 0)
        if res and res.goToFirstRow():
            return res.intFieldValue(0)
        return None

    def ps_helper_needs_installation(self):
        try:
            installed_version = get_installed_sys_version(self.main_view.editor)
            install_text = ""
            missing_grants = self.get_missing_grants()
            install_text = """\n\nTo be able to install the performance schema helper, you'll need the following privileges:
    - SELECT, INSERT, CREATE, DROP, ALTER, SUPER, CREATE VIEW, CREATE ROUTINE, ALTER ROUTINE and TRIGGER"""
            can_install = True
            if len(missing_grants) > 0:
                can_install = False
                install_text = "%s\n\nThe following grants are missing:\n  - %s" % (install_text, str(missing_grants))
    
            if not installed_version:
        
                return "The Performance Schema helper schema (sys) is not installed", \
    """Click the [Install Helper] button to install it.
    You must have at least the following privileges to use Performance Schema functionality:
      - SELECT on performance_schema.*
      - UPDATE on performance_schema.setup* for configuring instrumentation
      %s""" % install_text, can_install
            else:
                curversion = get_current_sys_version(self.ctrl_be.target_version)
                x, y, z = [int(i) for i in curversion.split(".")]
                ix, iy, iz = [int(i) for i in installed_version.split(".")]
                
                # major number is incremented on backwards incompatible changes
                # minor number is incremented when new things are added
                # release number is incremented on any other kind of change
                
                if ix == x and (iy > y or (iy == y and iz >= z)):
                    pass # ok
                elif x < ix:
                    return "Installed Performance Schema helper (sys) version is newer than supported", "MySQL Workbench needs to downgrade it.\n(current version is %s, server has %s).%s" % (curversion, installed_version, install_text), can_install
                else:
                    return "Performance Schema helper schema (sys) is outdated", "MySQL Workbench needs to upgrade it.\n(current version is %s, server has %s%s)." % (curversion, installed_version, install_text), can_install
        except grt.DBError, e:
            return "Unable to access Performance Schema helper schema (sys)", "%s (error %s)\n\nIf the sys schema is already installed, make sure you have SELECT privileges on it.\nIf not, you will need privileges to create the `sys` schema and populate it with views and stored procedures for PERFORMANCE_SCHEMA." % e.args

        except Exception, e:
            import traceback
            log_error("Error checking for PS helper: %s\n" % traceback.format_exc())
            return "Unable to access Performance Schema helper (sys)", str(e)+"\n\nIf the sys schema is already installed, make sure you have SELECT privileges on it.\nIf not, you will need privileges to create the `sys` schema and populate it with views and stored procedures for PERFORMANCE_SCHEMA."

        return None


    def ps_usable(self):
        """Checks if performance is enabled."""
        ret_val = False
        try:
            res = self.main_view.editor.executeManagementQuery("select @@performance_schema", 0)
            if res.goToFirstRow():
                value = res.stringFieldValue(0)
                ret_val = value == "1"
        except grt.DBError, e:
            log_error("MySQL error retrieving the performance_schema variable: %s\n" % e)
        
        return ret_val
    
    
    def check_usable(self):
        return None, None

    def get_missing_grants(self):
        # SELECT, INSERT, CREATE, DROP, ALTER, SUPER, CREATE VIEW, CREATE ROUTINE, ALTER ROUTINE, TRIGGER
        required_grants = ['SELECT', 'INSERT', 'CREATE', 'DROP', 'ALTER', 'SUPER', 'CREATE VIEW', 'CREATE ROUTINE', 'ALTER ROUTINE', 'TRIGGER']
        missing_grants = []
        current_user_grants = ""
        try:
            res = self.main_view.editor.executeManagementQuery("show grants", 0)
            if res.goToFirstRow():
                current_user_grants = res.stringFieldValue(0)
        except grt.DBError, e:
            log_error("MySQL error retrieving user grants: %s\n" % e)

        if current_user_grants.find('ALL') == -1:
            for grant in required_grants:
                if current_user_grants.find(grant) == -1:
                    missing_grants.append(grant)

        return missing_grants

    def page_activated(self):
        if self.warning_panel:
            self.remove(self.warning_panel)
            self.warning_panel = None

        self.page_active = True

        button_data = None
        if not self.ctrl_be.is_sql_connected():
            text = ("There is no connection to the MySQL Server.", "This functionality requires a connection to a MySQL server to work.", False)
        elif not self.ps_usable():
            text = ("Performance Schema Unavailable", "Performance Schema is either unavailable or disabled on this server.\nYou need a MySQL server version 5.6 or newer, with the performance_schema feature enabled.", False)
        else:
            text = self.ps_helper_needs_installation()
            
            if self.installer_panel:
                if self.installer_panel.is_busy:
                    return
                self.remove(self.installer_panel)
                self.installer_panel = None

            if text:
                button_data = ("Install Helper", self.install_helper)

        if not text:
            text, button_data = self.check_usable()

        if text:
            title, text, can_install = text
            if not can_install:
                button_data = None
            self.warning_panel = MessageButtonPanel("", title, text, button_data)
            self.add(self.warning_panel, True, True)
            
            if self.content:
                self.content.show(False)
        else:
            self.suspend_layout()
            if not self.ui_created:
                self.create_ui()
                self.ui_created = True
    
            self.content.show(True)
            self.relayout()
            self.resume_layout()


    def install_helper(self):
        self.remove(self.warning_panel)
        self.warning_panel = None

        self.installer_panel = HelperInstallPanel(self, self.main_view.editor)
        self.add(self.installer_panel, True, True)
        self.relayout() # needed b/c of layout bug in Mac

        self.installer_panel.importer.set_password(self.ctrl_be.get_mysql_password())
        self.installer_panel.start()





# Copyright (c) 2007, 2016, Oracle and/or its affiliates. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the
# License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301  USA

import os
import posixpath
import ntpath

write_log = False
logfile = "wbadebug.log"
debug_level = os.getenv("DEBUG_ADMIN")
if debug_level is not None:
    debug_level = int(debug_level)
    import inspect
else:
    debug_level = 0

if debug_level:
    print "Debug level -", debug_level

def dprint_ex(level, *args):
    if level <= debug_level:
        fr = inspect.currentframe().f_back
        cls = ""
        slf = fr.f_locals.get('self')
        if slf:
            cls = str(slf.__class__) + '.'
        ctx = inspect.getframeinfo(fr)
        # In Python 2.5, ctx is a tuple
        #method = cls + ctx.function + ':' + str(ctx.lineno)
        method = cls + ctx[2] + ":" + str(ctx[1])

        msg = method + " : " + " ".join([type(s) is str and s or str(s) for s in args])

        print msg
        if write_log:
            f = open(logfile, "a")
            f.write(msg)
            f.write("\n")
            f.close()



def splitpath(path):
    return posixpath.split(path) if '/' in path else ntpath.split(path)


def parentdir(path):
    return posixpath.dirname(path) if '/' in path else ntpath.dirname(path)


def stripdir(path):
    return posixpath.basename(path) if '/' in path else ntpath.basename(path)


def joinpath(path, *comps):
    return posixpath.join(path, *comps) if '/' in path else ntpath.join(path, *comps)


def sanitize_sudo_output(output):
    # in Mac OS X, XCode sets some DYLD_ environment variables when debugging which dyld
    # doesn't like and will print a warning to stderr, so we must filter that out
    if output.startswith("dyld: "):
        warning, _, output = output.partition("\n")
    return output

#===============================================================================
class Users:
  ADMIN = "root"
  CURRENT = ""

class OperationCancelledError(Exception):
    pass

# Put what is the wrong password in the exception message
class InvalidPasswordError(RuntimeError):
    pass

class PermissionDeniedError(RuntimeError):
    pass

class LogFileAccessError(RuntimeError):
    pass

class ServerIOError(RuntimeError):
    pass

class NoDriverInConnection(RuntimeError):
    pass

class SSHFingerprintNewError(Exception):
      def __init__(self, message, client, hostname, key):
          self.message = message
          self.key = key
          self.hostname = hostname
          self.client = client
          import binascii
          self.fingerprint = binascii.hexlify(self.key.get_fingerprint())
      def __str__(self):
         return self.message


def format_bad_host_exception(exc, known_hosts_filepath):
    import binascii
    return "The host %s fingerprints mismatch.\nExpected key: %s\nServer sent: %s\nPlease verify if it's correct.\nTo continue, delete entries for the host from the %s file." % (exc.hostname, binascii.hexlify(exc.expected_key.get_fingerprint()), binascii.hexlify(exc.key.get_fingerprint()), known_hosts_filepath)

# Decorator to log an exception
def log_error_decorator(method):
    def wrapper(self, error):
        import grt
        grt.log_error(self.__class__.__name__, str(error) + '\n')
        return method(self, error)
    return wrapper


class CmdOptions(object):
    CMD_WAIT_OUTPUT = 0
    CMD_HOME = 1
    
    def __setattr__(self, name, value):
        raise NotImplementedError

class CmdOutput(object):
    WAIT_ALWAYS = 1
    WAIT_NEVER = 2
    WAIT_IF_OK = 3
    WAIT_IF_FAIL = 4
    
    def __setattr__(self, name, value):
        raise NotImplementedError
